// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "math3d.h"

// vypocet TBN *************************************************************
void Cmath3d::calc_tbn(vec3 V0,vec3 V1,vec3 V2,vec2 V0T,vec2 V1T,vec2 V2T,
                       vec3 &T,vec3 &B,vec3 &N)
{
 // vypocet N
 v1v0=V1-V0; v2v0=V2-V0;
 N=v1v0.cross(v2v0); N.normalize();

 // vypocet T a B pre TANGENT space
 t1t0=V1T-V0T; t2t0=V2T-V0T;

 // recip=((t1.x-t0.x)*(t2.y-t0.y)-(t1.y-t0.y)*(t2.x-t0.x))
 recip=t1t0.x*t2t0.y-t1t0.y*t2t0.x;

 if (recip!=0.0f) {
  recip=1.0f/recip;
  //    (v1-v0).(t2-t0)y-(v2-v0).(t1-t0)y
  // T=-----------------------------------
  //                 recip
  T=(v1v0*t2t0.y-v2v0*t1t0.y)*recip; T.normalize();

  //    (v2-v0).(t1-t0)x-(v1-v0).(t2-t0)x
  // B=-----------------------------------
  //                 recip
  B=(v2v0*t1t0.x-v1v0*t2t0.x)*recip; B.normalize();
 }
/*
 // pre 3 vrcholy trojuholnika
 loopi(3) {
  const vec3 n=N[i];
  const vec3 t=T[i];
  float o;
  // Gram-Schmidt ortogonalizacia
  T[i]=(t-n*dot(n,t)).normalize();
  // lava/prava odchylka (handedness)
  o=(dot(cross(n,t),B[i])<0.0f) ? -1.0f : 1.0f;
 }*/
}

// nastavenie poloh vzoriek Gausovho filtra ********************************
void Cmath3d::set_gaussian_filter(int W,int H)
{
 if (W==0) W=1; if (H==0) H=1; // kontrola na delenie 0

 dx=1.0f/(float) W;
 horzTapOffs[0].set(0.0f,0.0f,       1.3366f*dx,0.0f);
 horzTapOffs[1].set(3.4295f*dx,0.0f, 5.4264f*dx,0.0f);
 horzTapOffs[2].set(7.4359f*dx,0.0f, 9.4436f*dx,0.0f);
 horzTapOffs[3].set(11.4401f*dx,0.0f,0.0f,0.0f);

 dy=1.0f/(float) H;
 vertTapOffs[0].set(0.0f,0.0f,       0.0f,1.3366f*dy);
 vertTapOffs[1].set(0.0f,3.4295f*dy, 0.0f,5.4264f*dy);
 vertTapOffs[2].set(0.0f,7.4359f*dy, 0.0f,9.4436f*dy);
 vertTapOffs[3].set(0.0f,11.4401f*dy,0.0f,0.0f);	
}

// quick-sort triedenie ****************************************************
void Cmath3d::quickSort(float *A,int *LI,int L,int R)
{
 int i=L,j=R,k;
 float pivot=A[((i+j)/2)*4+2];

 // porovnanie prvkov Z dvoch 4-prvkovych vektorov
 // v pripade potreby vymena XYZW prvkov dvoch vektorov so sucasnou vymenou indexov
 while (i<=j) {
  while (A[i*4+2]<pivot) i++;
  while (A[j*4+2]>pivot) j--;
  if (i<=j) {
   v.set(A[i*4+0],A[i*4+1],A[i*4+2],A[i*4+3]);
   A[i*4+0]=A[j*4+0]; A[i*4+1]=A[j*4+1]; A[i*4+2]=A[j*4+2]; A[i*4+3]=A[j*4+3];
   A[j*4+0]=v.x; A[j*4+1]=v.y; A[j*4+2]=v.z; A[j*4+3]=v.w;
   k=LI[i]; LI[i]=LI[j]; LI[j]=k;
   //printf("vymena %u %u ",i,j); loopk(16) printf("%.0f ",A[k*4+2]); printf("\n");
   i++; j--;
  }
 }

 // rekurzia
 if (L<j) quickSort(A,LI,L,j);
 if (i<R) quickSort(A,LI,i,R);
}
