#ifndef _SHADER_H_
#define _SHADER_H_

// praca so shadermi *******************************************************

// trieda premennych a funkcii pre pracu so shadermi
class CShader {
public:
 CShader();  // konstruktor
 ~CShader(); // destruktor

 // inicializacia pouzitia shaderov
 int init(void);

 // nacitanie kodu GLSL vertex & fragment shadera
 int load(const char *VP,const char *FP);
 // nacitanie kodu GLSL vertex & geometry & fragment shadera
 int load(const char *VP,const char *GP,const char *FP);
 int reload(const int T);
 void on(void);  // zapnutie GLSL shadera
 void off(void); // vypnutie GLSL shadera

 // nastavenie atributov
 void bindattriblocation(const GLuint I,const char *NAME);
 // nastavenie uniform premennych
 void uniformI(const char *NAME,const int I);
 void uniform1F(const char *NAME,const float F);
 void uniform2F(const char *NAME,const float F0,const float F1);
 void uniform3F(const char *NAME,const float F0,const float F1,const float F2);
 void uniformFV(const char *NAME,const int C,const GLfloat *V);
 void uniform4F(const char *NAME,const float F0,const float F1,const float F2,const float F3);
 void uniformmatrixFV(const char *NAME,const int C,const GLfloat *M);
private:
#ifdef OGL20
 GLuint vertex,geometry,fragment,program;
#else
 GLhandleARB vertex,geometry,fragment,program;
#endif

 char vp_name[32],gp_name[32],fp_name[32]; // nazov vertex, geometry, fragment programu

 char log[1000]; // logovanie pripadnych chyb pri preklade programov
 GLint log_length;

 GLint loc; // umiestnenie premennej v ramci programu
};

struct DOFsh {
 GLuint vertex,fragment,program;
};

// trieda premennych a funkcii pre pracu s Depth Of Field shadermi
class CDOFshaders {
public:
 CDOFshaders();  // konstruktor
 ~CDOFshaders(); // destruktor

 int init(void);    // inicializacia DOF shaderov
 void on(int P);    // zapnutie konkretneho DOF shadera
 void off(void);    // vypnutie DOF shaderov
 void cancel(void); // zrusenie DOF shaderov

 // nastavenie atributov
 void bindattriblocation(int P,const GLuint I,const char *NAME);
 // nastavenie uniform premennych
 void uniformI(int P,const char *NAME,const int I);
 void uniform1F(int P,const char *NAME,const float F);
 void uniformmatrixFV(int P,const char *NAME,const GLfloat *M);
 void uniform4FV(int P,const char *NAME,const GLfloat *V);
private:
 struct DOFsh sh[5];

 char log[1000]; // logovanie pripadnych chyb pri preklade programov
 GLint log_length;

 GLint loc; // umiestnenie premennej v ramci programu
};

/*
 odovzdavanie udajov:
 uniform - per primitive (outside glBegin - glEnd)
           hodnoty platia pre cely objekt, nie pre kazdy vrchol osobitne
 attrib - per vertex (inside glBegin - glEnd)
          pre kazdy vrchol osobitne, citat (nie menit) mozno iba vo vertex shaderi
 varying - interpolovane udaje prenasane z vertex do fragment shadera (po novom in, out)
 textury - udaje o farbach texelov mozu predstavovat odovzdavane udaje

 vertex shader - po vypocte per vertex atributov su tieto interpolovane a
                 odovzdane fragment shaderu
 fragment shader - vystupom operacii s interpolovanymi udajmi je farba fragmentu
*/

#endif
