// hlavickove subory *******************************************************
#include "includes.h"
// vlastny .h subor
#include "interface.h"

GLint glver=0,gl_max_texture_size=0,gl_max_texture_max_anisotropy_ext=0;

// uplynuty cas medzi aktualnym a povodnym casovym okamihom (s)
GLfloat dTime;

// pre pracu s matematikou
Cmath3d mm;

// instancia triedy CTexture - praca s texturami
CTexture texture;
// ciselne oznacenia textur - nie su totozne s identifikatormi nahratych textur v CTexture!
// vyhradene oznacenia:
// 0 - prazdna (empty) textura
// 1 - standardna (default) textura
// 2 - zameriavaci kriz
// 3-9 - specialne textury
// <10;(MAX_TEX_NUM-1)> - ostatne textury
GLuint tid[MAX_TEX_NUM];

// instancia triedy CText - praca s textom
CText text;

// pocet segmentov gule, valca, ihlana
int segs=18;
// pocet udajov / 3 = pocet trojuholnikov
/*
int data_num=108+                          // kocka
             (360/segs*18)*(180/segs)+     // gula
             (360/segs*18)+(360/segs*9)*2+ // valec
             (360/segs*9)+(360/segs*9)+    // ihlan
             18+18;                        // plochy
*/
// vrcholy, farba, TBN, texturovacie suradnice
// 108 + 3600 + 720 + 360 + 18 + 18 = 4824
// v pripade gule nie su generovane pary trojuholnikov na severnom
// a juznom pole, celkovy pocet je teda mensi o 40 trojuholnikov
float v_a[4824],c_a[4824],tan_a[4824],bit_a[4824],nor_a[4824],t_a[4824];
// pole indexov
// pocet trojuholnikov = 536 => pocet vrcholov = 536 * 3 = 1608
// pre triangles_adjacency je nutne kazdemu trojuholniku pridat 3 vrcholy
// pocet vrcholov = 536 * 6 = 3216
unsigned int i_a[3216];
unsigned int i_ac; // pocitadlo

// vektory totozne so suradnicovymi osami
vec3 rightV(1.0f,0.0f,0.0f);
vec3 upV(0.0f,1.0f,0.0f);
vec3 forwardV(0.0f,0.0f,-1.0f);

// v pripade sucasneho kreslenia do 4 color buffrov FBO
GLenum clrsbuf[]={GL_COLOR_ATTACHMENT0_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_COLOR_ATTACHMENT3_EXT};

// MSAA MultiSample AntiAliasing samples pre deferred shading
int msaa_samples=1;

// 1 gula=(360/segs*18)*(180/segs)=(360/30*18)*(180/30)=576 udajov=64 trojuholnikov
float sphere_v[576];
GLbuf *sph=NULL;

// konstruktor *************************************************************
CInterface::CInterface()
{
 ssao_samples=NULL;
}

// destruktor **************************************************************
CInterface::~CInterface()
{
 DELETEA(ssao_samples);

 // Vertex Array Object
 glDeleteVertexArrays(1,&objects_vao);
 glDeleteVertexArrays(1,&sphere_vao);
}

// nastavenie parametrov a udajov tlacidiel menu ***************************
void CInterface::set_buttons(void)
{
 // tlacidla menu
 if (but_text_id!=10 && but_text_id!=15) { // ak nie je HDR shader, alebo triangle_shrink
  but[0].init(vec3(10.0f,75.0f,0.0f),30.0f,15.0f); but[1].init(vec3(45.0f,75.0f,0.0f),30.0f,15.0f);
  but[2].init(vec3(80.0f,75.0f,0.0f),30.0f,15.0f); but[3].init(vec3(115.0f,75.0f,0.0f),30.0f,15.0f);

  but[4].init(vec3(10.0f,105.0f,0.0f),30.0f,15.0f); but[5].init(vec3(45.0f,105.0f,0.0f),30.0f,15.0f);
  but[6].init(vec3(80.0f,105.0f,0.0f),30.0f,15.0f); but[7].init(vec3(115.0f,105.0f,0.0f),30.0f,15.0f);

  but[8].init(vec3(10.0f,135.0f,0.0f),30.0f,15.0f); but[9].init(vec3(45.0f,135.0f,0.0f),30.0f,15.0f);
  but[10].init(vec3(80.0f,135.0f,0.0f),30.0f,15.0f); but[11].init(vec3(115.0f,135.0f,0.0f),30.0f,15.0f);

  but[12].init(vec3(10.0f,165.0f,0.0f),30.0f,15.0f); but[13].init(vec3(45.0f,165.0f,0.0f),30.0f,15.0f);
  but[14].init(vec3(80.0f,165.0f,0.0f),30.0f,15.0f); but[15].init(vec3(115.0f,165.0f,0.0f),30.0f,15.0f);

  but[16].init(vec3(10.0f,195.0f,0.0f),30.0f,15.0f); but[17].init(vec3(45.0f,195.0f,0.0f),30.0f,15.0f);
  but[18].init(vec3(80.0f,195.0f,0.0f),30.0f,15.0f);

  but[19].init(vec3(10.0f,225.0f,0.0f),36.0f,15.0f);
  but[20].init(vec3(10.0f,255.0f,0.0f),36.0f,15.0f);

  but[21].init(vec3(10.0f,285.0f,0.0f),30.0f,15.0f);
  but[22].init(vec3(10.0f,315.0f,0.0f),30.0f,15.0f);
  but[23].init(vec3(10.0f,345.0f,0.0f),30.0f,15.0f);

  but[24].init(vec3(10.0f,375.0f,0.0f),30.0f,15.0f); but[25].init(vec3(45.0f,375.0f,0.0f),30.0f,15.0f);
  but[26].init(vec3(80.0f,375.0f,0.0f),30.0f,15.0f); but[27].init(vec3(115.0f,375.0f,0.0f),30.0f,15.0f);

  but[28].init(vec3(10.0f,425.0f,0.0f),30.0f,15.0f); but[29].init(vec3(45.0f,425.0f,0.0f),30.0f,15.0f);
  but[30].init(vec3(80.0f,425.0f,0.0f),30.0f,15.0f); but[31].init(vec3(115.0f,425.0f,0.0f),30.0f,15.0f);

  but[32].init(vec3(10.0f,455.0f,0.0f),30.0f,15.0f); but[33].init(vec3(45.0f,455.0f,0.0f),30.0f,15.0f);
  but[34].init(vec3(80.0f,455.0f,0.0f),30.0f,15.0f); but[35].init(vec3(115.0f,455.0f,0.0f),30.0f,15.0f);

  but[36].init(vec3(10.0f,485.0f,0.0f),30.0f,15.0f); but[37].init(vec3(45.0f,485.0f,0.0f),30.0f,15.0f);
  but[38].init(vec3(80.0f,485.0f,0.0f),30.0f,15.0f); but[39].init(vec3(115.0f,485.0f,0.0f),30.0f,15.0f);

  but[40].init(vec3(10.0f,515.0f,0.0f),30.0f,15.0f); but[41].init(vec3(45.0f,515.0f,0.0f),30.0f,15.0f);
  but[42].init(vec3(80.0f,515.0f,0.0f),30.0f,15.0f); but[43].init(vec3(115.0f,515.0f,0.0f),30.0f,15.0f);

  but[44].init(vec3(10.0f,545.0f,0.0f),36.0f,15.0f);

  but[45].init(vec3(10.0f,575.0f,0.0f),135.0f,15.0f);

  but[46].init(vec3(95.0f,560.0f,0.0f),50.0f,14.0f);

  loopi(BUTNUM) but[i].set_tscale(0.3f,0.3f); // velkost textu tlacidiel

  // parametre a udaje tlacidiel
  but[0].set_data(lsrcparam.ambient.x); but[1].set_data(lsrcparam.ambient.y);
  but[2].set_data(lsrcparam.ambient.z); but[3].set_data(lsrcparam.ambient.w);

  but[4].set_data(lsrcparam.diffuse.x); but[5].set_data(lsrcparam.diffuse.y);
  but[6].set_data(lsrcparam.diffuse.z); but[7].set_data(lsrcparam.diffuse.w);

  but[8].set_data(lsrcparam.specular.x); but[9].set_data(lsrcparam.specular.y);
  but[10].set_data(lsrcparam.specular.z); but[11].set_data(lsrcparam.specular.w);

  but[12].set_range(-100.0f,100.0f); but[13].set_range(-100.0f,100.0f);
  but[14].set_range(-100.0f,100.0f);
  but[12].set_data(lsrcparam.position.x); but[13].set_data(lsrcparam.position.y);
  but[14].set_data(lsrcparam.position.z); but[15].set_data(lsrcparam.position.w);

  but[16].set_range(-1.0f,1.0f); but[17].set_range(-1.0f,1.0f);
  but[18].set_range(-1.0f,1.0f);
  but[16].set_data(lsrcparam.spotDirection.x); but[17].set_data(lsrcparam.spotDirection.y);
  but[18].set_data(lsrcparam.spotDirection.z);

  but[19].set_range(0.0f,128.0f); but[19].set_cnum(5);
  but[19].set_data(lsrcparam.spotExponent);
  but[20].set_range(0.0f,90.0f); but[20].set_cnum(5);
  but[20].set_data(lsrcparam.spotCutoff);

  but[21].set_data(lsrcparam.constantAttenuation);
  but[22].set_data(lsrcparam.linearAttenuation);
  but[23].set_data(lsrcparam.quadraticAttenuation);

  but[24].set_data(lmodelparam.ambient.x); but[25].set_data(lmodelparam.ambient.y);
  but[26].set_data(lmodelparam.ambient.z); but[27].set_data(lmodelparam.ambient.w);

  but[28].set_data(mparam.ambient.x); but[29].set_data(mparam.ambient.y);
  but[30].set_data(mparam.ambient.z); but[31].set_data(mparam.ambient.w);

  but[32].set_data(mparam.diffuse.x); but[33].set_data(mparam.diffuse.y);
  but[34].set_data(mparam.diffuse.z); but[35].set_data(mparam.diffuse.w);

  but[36].set_data(mparam.specular.x); but[37].set_data(mparam.specular.y);
  but[38].set_data(mparam.specular.z); but[39].set_data(mparam.specular.w);

  but[40].set_data(mparam.emission.x); but[41].set_data(mparam.emission.y);
  but[42].set_data(mparam.emission.z); but[43].set_data(mparam.emission.w);

  but[44].set_range(0.0f,128.0f); but[44].set_cnum(5);
  but[44].set_data(mparam.shininess);
 }
 else if (but_text_id==10) {
  // pre HDR shader, tlacidla so sirkou 0 sa nezobrazuju, inicializovane
  // su iba z dovodu zosuladenia oboch casti algoritmu a zjednodusenia prace
  but[0].init(vec3(10.0f,75.0f,0.0f),30.0f,15.0f); but[1].init(vec3(45.0f,75.0f,0.0f),0.0f,0.0f);
  but[2].init(vec3(80.0f,75.0f,0.0f),0.0f,0.0f); but[3].init(vec3(115.0f,75.0f,0.0f),0.0f,0.0f);

  but[4].init(vec3(10.0f,105.0f,0.0f),30.0f,15.0f); but[5].init(vec3(45.0f,105.0f,0.0f),0.0f,0.0f);
  but[6].init(vec3(80.0f,105.0f,0.0f),0.0f,0.0f); but[7].init(vec3(115.0f,105.0f,0.0f),0.0f,0.0f);

  but[8].init(vec3(10.0f,135.0f,0.0f),30.0f,15.0f); but[9].init(vec3(45.0f,135.0f,0.0f),0.0f,0.0f);
  but[10].init(vec3(80.0f,135.0f,0.0f),0.0f,0.0f); but[11].init(vec3(115.0f,135.0f,0.0f),0.0f,0.0f);

  but[12].init(vec3(10.0f,185.0f,0.0f),30.0f,15.0f); but[13].init(vec3(45.0f,165.0f,0.0f),0.0f,0.0f);
  but[14].init(vec3(80.0f,165.0f,0.0f),0.0f,0.0f); but[15].init(vec3(115.0f,165.0f,0.0f),0.0f,0.0f);

  but[16].init(vec3(10.0f,215.0f,0.0f),30.0f,15.0f); but[17].init(vec3(45.0f,195.0f,0.0f),0.0f,0.0f);
  but[18].init(vec3(80.0f,195.0f,0.0f),0.0f,0.0f);

  but[19].init(vec3(10.0f,245.0f,0.0f),30.0f,15.0f);
  but[20].init(vec3(10.0f,255.0f,0.0f),0.0f,0.0f);

  but[21].init(vec3(10.0f,295.0f,0.0f),30.0f,15.0f);
  but[22].init(vec3(10.0f,325.0f,0.0f),30.0f,15.0f);
  but[23].init(vec3(10.0f,355.0f,0.0f),30.0f,15.0f);

  but[24].init(vec3(10.0f,375.0f,0.0f),0.0f,0.0f); but[25].init(vec3(45.0f,375.0f,0.0f),0.0f,0.0f);
  but[26].init(vec3(80.0f,375.0f,0.0f),0.0f,0.0f); but[27].init(vec3(115.0f,375.0f,0.0f),0.0f,0.0f);

  but[28].init(vec3(10.0f,425.0f,0.0f),0.0f,0.0f); but[29].init(vec3(45.0f,425.0f,0.0f),0.0f,0.0f);
  but[30].init(vec3(80.0f,425.0f,0.0f),0.0f,0.0f); but[31].init(vec3(115.0f,425.0f,0.0f),0.0f,0.0f);

  but[32].init(vec3(10.0f,455.0f,0.0f),0.0f,0.0f); but[33].init(vec3(45.0f,455.0f,0.0f),0.0f,0.0f);
  but[34].init(vec3(80.0f,455.0f,0.0f),0.0f,0.0f); but[35].init(vec3(115.0f,455.0f,0.0f),0.0f,0.0f);

  but[36].init(vec3(10.0f,485.0f,0.0f),0.0f,0.0f); but[37].init(vec3(45.0f,485.0f,0.0f),0.0f,0.0f);
  but[38].init(vec3(80.0f,485.0f,0.0f),0.0f,0.0f); but[39].init(vec3(115.0f,485.0f,0.0f),0.0f,0.0f);

  but[40].init(vec3(10.0f,515.0f,0.0f),0.0f,0.0f); but[41].init(vec3(45.0f,515.0f,0.0f),0.0f,0.0f);
  but[42].init(vec3(80.0f,515.0f,0.0f),0.0f,0.0f); but[43].init(vec3(115.0f,515.0f,0.0f),0.0f,0.0f);

  but[44].init(vec3(10.0f,545.0f,0.0f),0.0f,0.0f);

  but[45].init(vec3(10.0f,575.0f,0.0f),135.0f,15.0f);

  but[46].init(vec3(95.0f,560.0f,0.0f),50.0f,14.0f);

  loopi(BUTNUM) but[i].set_tscale(0.3f,0.3f); // velkost textu tlacidiel

  // parametre a udaje tlacidiel
  but[0].set_data(hdrparam.refract.x);
  but[4].set_data(hdrparam.refract.y);
  but[8].set_data(hdrparam.refract.z);

  but[12].set_data(hdrparam.fresnel.x);
  but[16].set_data(hdrparam.fresnel.y);
  but[19].set_data(hdrparam.fresnel.z);

  but[21].set_data(hdrparam.bloom.x);
  but[22].set_range(0.0f,2.0f); but[22].set_data(hdrparam.bloom.y);
  but[23].set_range(0.0f,2.0f); but[23].set_data(hdrparam.bloom.z);
 }
 else {
  // pre triangle_shrink, tlacidla so sirkou 0 sa nezobrazuju, inicializovane
  // su iba z dovodu zosuladenia oboch casti algoritmu a zjednodusenia prace
  but[0].init(vec3(10.0f,75.0f,0.0f),30.0f,15.0f); but[1].init(vec3(45.0f,75.0f,0.0f),0.0f,0.0f);
  but[2].init(vec3(80.0f,75.0f,0.0f),0.0f,0.0f); but[3].init(vec3(115.0f,75.0f,0.0f),0.0f,0.0f);

  but[4].init(vec3(10.0f,105.0f,0.0f),30.0f,15.0f); but[5].init(vec3(45.0f,105.0f,0.0f),0.0f,0.0f);
  but[6].init(vec3(80.0f,105.0f,0.0f),0.0f,0.0f); but[7].init(vec3(115.0f,105.0f,0.0f),0.0f,0.0f);

  but[8].init(vec3(10.0f,135.0f,0.0f),0.0f,0.0f); but[9].init(vec3(45.0f,135.0f,0.0f),0.0f,0.0f);
  but[10].init(vec3(80.0f,135.0f,0.0f),0.0f,0.0f); but[11].init(vec3(115.0f,135.0f,0.0f),0.0f,0.0f);

  but[12].init(vec3(10.0f,185.0f,0.0f),0.0f,0.0f); but[13].init(vec3(45.0f,165.0f,0.0f),0.0f,0.0f);
  but[14].init(vec3(80.0f,165.0f,0.0f),0.0f,0.0f); but[15].init(vec3(115.0f,165.0f,0.0f),0.0f,0.0f);

  but[16].init(vec3(10.0f,215.0f,0.0f),0.0f,0.0f); but[17].init(vec3(45.0f,195.0f,0.0f),0.0f,0.0f);
  but[18].init(vec3(80.0f,195.0f,0.0f),0.0f,0.0f);

  but[19].init(vec3(10.0f,245.0f,0.0f),0.0f,0.0f);
  but[20].init(vec3(10.0f,255.0f,0.0f),0.0f,0.0f);

  but[21].init(vec3(10.0f,295.0f,0.0f),0.0f,0.0f);
  but[22].init(vec3(10.0f,325.0f,0.0f),0.0f,0.0f);
  but[23].init(vec3(10.0f,355.0f,0.0f),0.0f,0.0f);

  but[24].init(vec3(10.0f,375.0f,0.0f),0.0f,0.0f); but[25].init(vec3(45.0f,375.0f,0.0f),0.0f,0.0f);
  but[26].init(vec3(80.0f,375.0f,0.0f),0.0f,0.0f); but[27].init(vec3(115.0f,375.0f,0.0f),0.0f,0.0f);

  but[28].init(vec3(10.0f,425.0f,0.0f),0.0f,0.0f); but[29].init(vec3(45.0f,425.0f,0.0f),0.0f,0.0f);
  but[30].init(vec3(80.0f,425.0f,0.0f),0.0f,0.0f); but[31].init(vec3(115.0f,425.0f,0.0f),0.0f,0.0f);

  but[32].init(vec3(10.0f,455.0f,0.0f),0.0f,0.0f); but[33].init(vec3(45.0f,455.0f,0.0f),0.0f,0.0f);
  but[34].init(vec3(80.0f,455.0f,0.0f),0.0f,0.0f); but[35].init(vec3(115.0f,455.0f,0.0f),0.0f,0.0f);

  but[36].init(vec3(10.0f,485.0f,0.0f),0.0f,0.0f); but[37].init(vec3(45.0f,485.0f,0.0f),0.0f,0.0f);
  but[38].init(vec3(80.0f,485.0f,0.0f),0.0f,0.0f); but[39].init(vec3(115.0f,485.0f,0.0f),0.0f,0.0f);

  but[40].init(vec3(10.0f,515.0f,0.0f),0.0f,0.0f); but[41].init(vec3(45.0f,515.0f,0.0f),0.0f,0.0f);
  but[42].init(vec3(80.0f,515.0f,0.0f),0.0f,0.0f); but[43].init(vec3(115.0f,515.0f,0.0f),0.0f,0.0f);

  but[44].init(vec3(10.0f,545.0f,0.0f),0.0f,0.0f);

  but[45].init(vec3(10.0f,575.0f,0.0f),135.0f,15.0f);

  but[46].init(vec3(95.0f,560.0f,0.0f),50.0f,14.0f);

  loopi(BUTNUM) but[i].set_tscale(0.3f,0.3f); // velkost textu tlacidiel

  // parametre a udaje tlacidiel
  but[0].set_data(shrink_level);
  but[4].set_data(rotate_speed);
 }
}

// inicializacia OpenGL - nastavenie parametrov OpenGL sceny ***************
int CInterface::OGL_initialize(void)
{
 printf("OpenGL init...\n");

 GLenum err=glewInit();
 if (err!=GLEW_OK) { printf("GLEW error: %s\n",glewGetErrorString(err)); return(1); }
 printf("using GLEW %s\n",glewGetString(GLEW_VERSION));

 // informacia o verzii OpenGL
 int glmaj,glmin;
 getglver(&glmaj,&glmin);
 glver=glmaj*10+glmin;
 printf("using OpenGL %d.%d\n",glmaj,glmin);
 if (glmaj<2) {
  printf("Pre spustenie tejto aplikacie je potrebna kniznica OpenGL 2.0!\n");
  return(1);
 }
 printf("using GLSL %s\n",getglslver());
 val=glGetString(GL_VENDOR); printf("vendor: %s ",val);
 val=glGetString(GL_RENDERER); printf("renderer: %s\n",val);

 if (!GLEW_ARB_texture_non_power_of_two) {
  printf("HW nepodporuje rozsirenie GL_ARB_texture_non_power_of_two!\n");
 }

 glGetIntegerv(GL_MAX_TEXTURE_SIZE,&gl_max_texture_size);
 printf("gl_max_texture_size: %d\n",gl_max_texture_size);
 if (GLEW_EXT_texture_filter_anisotropic) {
  glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT,&gl_max_texture_max_anisotropy_ext);
  printf("gl_max_texture_max_anisotropy_ext: %d\n",gl_max_texture_max_anisotropy_ext);
 }

 if (!GLEW_ARB_framebuffer_object) {
  printf("HW nepodporuje rozsirenie GL_ARB_framebuffer_object!\n");
  return(1);
 }

 // STAVOVY STROJ STAVOVY STROJ STAVOVY STROJ STAVOVY STROJ STAVOVY STROJ --

 // farba, ktorou sa bude resetovat zasobnik farieb
 glClearColor(0.0f,0.0f,0.0f,1.0f);
 // zasobnik hlbky bude vymazavany nastavenim na 1.0f
 glClearDepth(1.0f);
 // pomocny zasobnik bude vymazavany nastavenim na 0.0f
 glClearStencil(0.0f);

 glDepthFunc(GL_LESS);    // typ porovnavacej funkcie pre urcenie Z-vzdialenosti
 glEnable(GL_DEPTH_TEST); // testovanie Z-vzdialenosti

 glDisable(GL_CULL_FACE); // poradie, v ktorom budu vrcholy objektov
 glFrontFace(GL_CCW);     // definovat privratene, resp. odvratene plochy,
 glCullFace(GL_BACK);     // CCW - proti smeru hod. ruciciek, CW - v smere hod. ruciciek
 glEnable(GL_CULL_FACE);  // odvratene plochy sa nebudu spracovavat

 glEnable(GL_TEXTURE_2D); // texturovanie je standardne zapnute

 //glPolygonMode(GL_FRONT,GL_LINE);

 //glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
 //glHint(GL_GENERATE_MIPMAP_HINT,GL_NICEST);

 // osvetlenie
 lsrcparam.ambient.set(0.9f,0.9f,0.9f,1.0f);
 lsrcparam.diffuse.set(0.6f,1.0f,1.0f,1.0f);
 lsrcparam.specular.set(1.0f,1.0f,1.0f,1.0f);
 lsrcparam.position.set(-2.0f,11.0f,15.0f,1.0f);
 lsrcparam.spotDirection.set(0.0f,0.0f,-1.0f);
 lsrcparam.spotExponent=30.0f;
 lsrcparam.spotCutoff=30.0f;
 lsrcparam.spotCosCutoff=cos(DEG2RAD(lsrcparam.spotCutoff));
 lsrcparam.spotSinCutoff=sin(DEG2RAD(lsrcparam.spotCutoff));
 lsrcparam.constantAttenuation=0.3f;
 lsrcparam.linearAttenuation=0.05f;
 lsrcparam.quadraticAttenuation=0.01f;

 lmodelparam.ambient.set(0.25f,0.25f,0.25f,1.0f);

 mparam.ambient.set(0.6f,0.6f,0.6f,1.0f);
 mparam.diffuse.set(0.8f,0.4f,0.4f,1.0f);
 mparam.specular.set(1.0f,1.0f,1.0f,1.0f);
 mparam.emission.set(0.0f,0.0f,0.0f,1.0f);
 mparam.shininess=128.0f;

 srand((int) SDL_GetTicks());
 // 16 svetelnych zdrojov
 loopi(16) {
  mlight[i].diffuse.set((float) rand()/(float) RAND_MAX,
                        (float) rand()/(float) RAND_MAX,
                        (float) rand()/(float) RAND_MAX,1.0);
  mlight[i].position.set(0.0f,0.0f,0.0f,1.0f);
  mlight[i].angle.set(0.0f,0.0f,0.0f);
  mlight[i].omega.set((float) rand()/(float) RAND_MAX*PI_MUL_2/400.0f,
                      (float) rand()/(float) RAND_MAX*PI_MUL_2/400.0f,
                      (float) rand()/(float) RAND_MAX*PI_MUL_2/400.0f);
 }

 // HDR parametre
 hdrparam.refract.set(0.6f,0.5f,0.4f);                  // red, green, blue
 hdrparam.fresnel.set(0.7f,0.8f,0.1f);                  // bias, scale, power
 hdrparam.bloom.set(0.8f,1.2f,(exp(0.86f*0.86f)-1.0f)); // factor, bright, exposure

 // aktualizacia parametrov osvetlenia
 update_lighting_params();

 // INICIALIZACIA SCENY INICIALIZACIA SCENY INICIALIZACIA SCENY ------------

 // nastavenie ciselnych oznaceni textur
 // vyhradene oznacenia:
 // 0 - prazdna (empty) textura
 // 1 - standardna (default) textura
 // 2 - zameriavaci kriz
 // 3-9 - specialne textury
 // <10;(MAX_TEX_NUM-1)> - ostatne textury
 loopi(MAX_TEX_NUM) tid[i]=i;

 // inicializacia Avatara
 avatar.init();

 // inicializacia textu - nahratie textury s fontom
 if (text.init("data/font/font.png")) {
  printf("Chyba inicializacie textu!\n"); return(1);
 }

 // nahratie prazdnej (empty) textury
 if (texture.load2D(tid[0],"data/textures/default/empty.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia prazdnej textury!\n"); return(1);
 }
 // nahratie standardnej (default) textury
 if (texture.load2D(tid[1],"data/textures/default/default.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia standardnej textury!\n"); return(1);
 }
 // nahratie textury zameriavacieho kriza
 if (texture.load2D(tid[2],"data/gfx/crosshair1.png",T_LINEAR | T_CLAMP)) {
  printf("Chyba nahratia textury zameriavacieho kriza!\n"); return(1);
 }
 // priprava 3D textury
 if (texture.prepare_perlin_noise_tex(tid[3])) {
  printf("Chyba pripravy 3D textury!\n"); return(1);
 }
 // priprava 2D textury nahodne orientovanych vektorov
 if (texture.prepare_random_vec_tex(tid[4])) {
  printf("Chyba pripravy 2D textury!\n"); return(1);
 }

 // nahratie textury predstavujucej svetelny zdroj
 if (texture.load2D(tid[10],"data/textures/lamp.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury svetelneho zdroja!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[11],"data/textures/chessboard.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }
 // nahratie normalmap textury
 if (texture.load2D(tid[12],"data/textures/chessboard_nm.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[13],"data/textures/px3/color_map.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }
 // nahratie heightmap textury
 if (texture.load2D(tid[14],"data/textures/px3/height_map.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }
 // nahratie normalmap textury
 if (texture.load2D(tid[15],"data/textures/px3/normal_map.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[16],"data/textures/ceramic.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[17],"data/textures/desert.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[18],"data/textures/ChewingGumFactory.jpeg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury predstavujucej svetelny zdroj
 if (texture.load2D(tid[19],"data/gfx/light_bulb.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury svetelneho zdroja!\n"); return(1);
 }

 // nahratie 6-tich textur - identifikatory 20-25
 // texura s tid[26] je konkretna CubeMap textura
 if (texture.loadCubeMap(tid[20],"data/textures/env_cross.bmp")) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie textury
 if (texture.load2D(tid[27],"data/textures/graychess.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }
 // nahratie normalmap textury
 if (texture.load2D(tid[28],"data/textures/graychess_nm.jpg",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 if (texture.load2D(tid[29],"data/textures/SwirlySlateTile.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }
 if (texture.load2D(tid[30],"data/textures/SwirlySlateTile_nm.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia textury!\n"); return(1);
 }

 // nahratie white textury
 if (texture.load2D(tid[31],"data/textures/default/white.png",T_LINEAR | T_MIPMAP)) {
  printf("Chyba nahratia white textury!\n"); return(1);
 }

 // velkost zameriavacieho kriza
 crosshairSize=50.0f;

 // uroven zuzenia pre triangle_shrink
 shrink_level=0.6f;
 // uhol rotacie
 rotate_angle=0.0f;
 // rychlost rotacie
 rotate_speed=0.3f;

 but_text_id=SHNUM-1; // index textoveho udaju tlacidla = pouzity shader

 set_buttons(); // nastavenie parametrov a udajov tlacidiel menu

 strcpy(but_text[0],"base");  // SHNUM x textovy udaj tlacidla = nazvy shaderov
 strcpy(but_text[1],"light");
 strcpy(but_text[2],"brick");
 strcpy(but_text[3],"toon");
 strcpy(but_text[4],"wave");
 strcpy(but_text[5],"wood");
 strcpy(but_text[6],"normalmap");
 strcpy(but_text[7],"parallax");
 strcpy(but_text[8],"water");
 strcpy(but_text[9],"dof");
 strcpy(but_text[10],"hdr");
 strcpy(but_text[11],"ssao");
 strcpy(but_text[12],"shadows");
 strcpy(but_text[13],"deferred");
 strcpy(but_text[14],"mblur");
 strcpy(but_text[15],"trishrink");
 strcpy(but_text[16],"silhouette");
 but[45].set_range(0.0f,16.1f);                       // po konverzii na int bude ciselny udaj 0-16
 but[45].set_show_data(1);                            // tlacidlo bude zobrazovat textovy udaj
 but[45].set_data(but_text_id,but_text[but_text_id]); // ciselny a textovy udaj tlacidla

 but[46].set_show_data(1);     // tlacidlo bude zobrazovat textovy udaj
 but[46].set_data(0,"reload"); // ciselny a textovy udaj tlacidla

 but_id=-1; // index aktivneho tlacidla

 // FBO
 fbo.init();
 hdr_fbo.init0();
 downsampler_fbo.init_downsampler();
 ssao_fbo.init_ssao();
 shadowmap_fbo.init_shadowmap();
 raw_fbo.init_raw();

 // priprava objektov - kocka, gula, valec, ihlan, plochy
 objects_prepare();
 // priprava objektu - gula reprezentujuca svetelny zdroj
 prepare_light_sphere(3.8f);

 // inicializacia a nacitanie shaderov
 if (base_shader.init()) return(1);
 if (base_shader.load("shaders/base.vp","shaders/base.fp")) return(1);
 if (light_shader.load("shaders/light.vp","shaders/light.fp")) return(1);
 if (brick_shader.load("shaders/brick.vp","shaders/brick.fp")) return(1);
 if (toon_shader.load("shaders/toon.vp","shaders/toon.fp")) return(1);
 if (wave_shader.load("shaders/wave.vp","shaders/wave.fp")) return(1);
 if (wood_shader.load("shaders/wood.vp","shaders/wood.fp")) return(1);
 if (nm_shader.load("shaders/normalmap.vp","shaders/normalmap.fp")) return(1);
 if (px_shader.load("shaders/parallax.vp","shaders/parallax.fp")) return(1);
 if (water_shader.load("shaders/water.vp","shaders/water.fp")) return(1);
 if (dof_shader.init()) return(1);
 if (chromatic_shader.load("shaders/chromatic.vp","shaders/chromatic.fp")) return(1);
 if (bloom_shader.load("shaders/bloom.vp","shaders/bloom.fp")) return(1);
 if (blur_shader.load("shaders/blur.vp","shaders/blur.fp")) return(1);
 if (hdr_shader.load("shaders/hdr.vp","shaders/hdr.fp")) return(1);
 if (ssao_shader[0].load("shaders/ssao/ssao_shader0.vp","shaders/ssao/ssao_shader0.fp")) return(1);
 if (ssao_shader[1].load("shaders/ssao/ssao_shader1.vp","shaders/ssao/ssao_shader1.fp")) return(1);
 if (ssao_shader[2].load("shaders/ssao/ssao_shader2.vp","shaders/ssao/ssao_shader2.fp")) return(1);
 if (ssao_shader[3].load("shaders/ssao/ssao_shader3.vp","shaders/ssao/ssao_shader3.fp")) return(1);
 if (ssao_shader[4].load("shaders/ssao/ssao_shader4.vp","shaders/ssao/ssao_shader4.fp")) return(1);
 if (shadow_shader.load("shaders/shadow.vp","shaders/shadow.fp")) return(1);
 if (null_shader.load("shaders/null.vp","shaders/null.fp")) return(1);
 if (raw_shader.load("shaders/raw.vp","shaders/raw.fp")) return(1);
 if (def_shader.load("shaders/def.vp","shaders/def.fp")) return(1);
 if (msaa_shader.load("shaders/msaa.vp","shaders/msaa.fp")) return(1);
 if (deltas_shader.load("shaders/deltas.vp","shaders/deltas.fp")) return(1);
 if (mblur_shader.load("shaders/mblur.vp","shaders/mblur.fp")) return(1);
 if (triangle_shrink.load("shaders/light2.vp","shaders/trishrink.gp","shaders/light2.fp")) return(1);
 if (silhouette_shader.load("shaders/silhouette.vp","shaders/silhouette.gp","shaders/silhouette.fp")) return(1);

 base_shader.on();
 base_shader.bindattriblocation(0,"in_pos"); // vrcholy
 base_shader.bindattriblocation(1,"in_col"); // farby
 base_shader.bindattriblocation(4,"in_nor"); // normaly
 base_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 base_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 base_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 base_shader.off();

 light_shader.on();
 light_shader.bindattriblocation(0,"in_pos"); // vrcholy
 light_shader.bindattriblocation(1,"in_col"); // farby
 light_shader.bindattriblocation(4,"in_nor"); // normaly
 light_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 light_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 light_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 light_shader.off();

 brick_shader.on();
 brick_shader.bindattriblocation(0,"in_pos"); // vrcholy
 brick_shader.bindattriblocation(4,"in_nor"); // normaly
 brick_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 brick_shader.off();

 toon_shader.on();
 toon_shader.bindattriblocation(0,"in_pos"); // vrcholy
 toon_shader.bindattriblocation(4,"in_nor"); // normaly
 toon_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 toon_shader.off();

 wave_shader.on();
 wave_shader.bindattriblocation(0,"in_pos"); // vrcholy
 wave_shader.bindattriblocation(1,"in_col"); // farby
 wave_shader.bindattriblocation(4,"in_nor"); // normaly
 wave_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 wave_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 wave_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 wave_shader.off();

 wood_shader.on();
 wood_shader.bindattriblocation(0,"in_pos"); // vrcholy
 wood_shader.bindattriblocation(4,"in_nor"); // normaly
 wood_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 wood_shader.uniform3F("lightWoodColor",0.6f,0.3f,0.1f);
 wood_shader.uniform3F("darkWoodColor",0.4f,0.2f,0.07f);
 wood_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 wood_shader.off();

 nm_shader.on();
 nm_shader.bindattriblocation(0,"in_pos"); // vrcholy
 nm_shader.bindattriblocation(2,"in_tan"); // tangenty
 nm_shader.bindattriblocation(3,"in_bit"); // bitangenty
 nm_shader.bindattriblocation(4,"in_nor"); // normaly
 nm_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 nm_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 nm_shader.uniformI("nm_tex_unit",1);      // normalmap texturovacia jednotka
 nm_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 nm_shader.off();

 px_shader.on();
 px_shader.bindattriblocation(0,"in_pos"); // vrcholy
 px_shader.bindattriblocation(2,"in_tan"); // tangenty
 px_shader.bindattriblocation(3,"in_bit"); // bitangenty
 px_shader.bindattriblocation(4,"in_nor"); // normaly
 px_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 px_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 px_shader.uniformI("hm_tex_unit",1);      // heightmap texturovacia jednotka
 px_shader.uniformI("nm_tex_unit",2);      // normalmap texturovacia jednotka
 px_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 px_shader.off();

 // inicializacia objektu predstavujuceho hladinu vody
 if (wtr.init()) return(1);

 water_shader.on();
 water_shader.bindattriblocation(0,"in_pos"); // vrcholy
 water_shader.bindattriblocation(2,"in_tan"); // tangenty
 water_shader.bindattriblocation(3,"in_bit"); // bitangenty
 water_shader.bindattriblocation(4,"in_nor"); // normaly
 water_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 water_shader.uniformI("rfl_tex_unit",0);     // reflection texturovacia jednotka
 water_shader.uniformI("rfr_tex_unit",1);     // refraction texturovacia jednotka
 water_shader.uniformI("depth_tex_unit",2);   // depth texturovacia jednotka
 water_shader.uniformI("nm_tex_unit",3);      // normalmap texturovacia jednotka
 water_shader.uniformI("dudv_tex_unit",4);    // dudv texturovacia jednotka
 water_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 water_shader.off();

 // depth of field shadery - na efekte spolupracuje celkovo 5 shaderov
 dof_shader.on(0);
 dof_shader.bindattriblocation(0,0,"in_pos");
 dof_shader.bindattriblocation(0,8,"in_uv");
 dof_shader.uniformmatrixFV(0,"projection_mat",avatar.get_pr_mat());
 dof_shader.uniformI(0,"base_tex_unit",0);
 dof_shader.uniform1F(0,"focalDistance",3.0f);
 dof_shader.uniform1F(0,"focalRange",6.0f);

 dof_shader.on(1);
 dof_shader.bindattriblocation(1,0,"in_pos");
 dof_shader.bindattriblocation(1,8,"in_uv");
 dof_shader.uniformI(1,"base_tex_unit",0);

 // nastavenie poloh vzoriek Gausovho filtra
 mm.set_gaussian_filter(scr_w*2,scr_h*2);
 
 dof_shader.on(2);
 dof_shader.bindattriblocation(2,0,"in_pos");
 dof_shader.bindattriblocation(2,8,"in_uv");
 dof_shader.uniformI(2,"base_tex_unit",0);
 dof_shader.uniform4FV(2,"horzTapOffs0",mm.horzTapOffs[0].get_data());
 dof_shader.uniform4FV(2,"horzTapOffs1",mm.horzTapOffs[1].get_data());
 dof_shader.uniform4FV(2,"horzTapOffs2",mm.horzTapOffs[2].get_data());
 dof_shader.uniform4FV(2,"horzTapOffs3",mm.horzTapOffs[3].get_data());

 dof_shader.on(3);
 dof_shader.bindattriblocation(3,0,"in_pos");
 dof_shader.bindattriblocation(3,8,"in_uv");
 dof_shader.uniformI(3,"base_tex_unit",0);
 dof_shader.uniform4FV(3,"vertTapOffs0",mm.vertTapOffs[0].get_data());
 dof_shader.uniform4FV(3,"vertTapOffs1",mm.vertTapOffs[1].get_data());
 dof_shader.uniform4FV(3,"vertTapOffs2",mm.vertTapOffs[2].get_data());
 dof_shader.uniform4FV(3,"vertTapOffs3",mm.vertTapOffs[3].get_data());

 dof_shader.on(4);
 dof_shader.bindattriblocation(4,0,"in_pos");
 dof_shader.bindattriblocation(4,8,"in_uv");
 dof_shader.uniformI(4,"base_tex_unit",0);
 dof_shader.uniformI(4,"blur_tex_unit",1);

 dof_shader.off();

 // inicializacia objektu predstavujuceho okolite prostredie
 if (enviro.init()) return(1);

 chromatic_shader.on();
 chromatic_shader.bindattriblocation(0,"in_pos"); // vrcholy
 chromatic_shader.bindattriblocation(1,"in_col"); // farby
 chromatic_shader.bindattriblocation(4,"in_nor"); // normaly
 chromatic_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 chromatic_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 chromatic_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 chromatic_shader.uniformFV("refract",1,hdrparam.refract.get_data()); // refract R, G, B
 chromatic_shader.uniformFV("fresnel",1,hdrparam.fresnel.get_data()); // bias, scale, power
 chromatic_shader.off();

 bloom_shader.on();
 bloom_shader.bindattriblocation(0,"in_pos"); // vrcholy
 bloom_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 bloom_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 bloom_shader.off();

 blur_shader.on();
 blur_shader.bindattriblocation(0,"in_pos"); // vrcholy
 blur_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 blur_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 blur_shader.off();

 hdr_shader.on();
 hdr_shader.bindattriblocation(0,"in_pos"); // vrcholy
 hdr_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 hdr_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 hdr_shader.uniformI("bloom_tex_unit",1);   // bloom texturovacia jednotka
 hdr_shader.uniformFV("bloom",1,hdrparam.bloom.get_data()); // factor, bright, exposure
 hdr_shader.off();

 ssao_shader[0].on();
 ssao_shader[0].bindattriblocation(0,"in_pos"); // vrcholy
 ssao_shader[0].bindattriblocation(4,"in_nor"); // normaly
 ssao_shader[0].bindattriblocation(8,"in_uv");  // texturovacie suradnice
 ssao_shader[0].uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 ssao_shader[0].uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 ssao_shader[0].off();

 // 16 ssao vzoriek, kazda ma 3 float hodnoty X,Y,Z, vzorky nie su normalizovane
 ssao_samples=new float[16*3];
 float *ptr=ssao_samples; // pomocny ukazovatel
 float alfa=PI_DIV_4,beta=PI_DIV_2/8.0f,radius=0.3f;
 int j=0; vec3 v;
 loopi(16) {
  v.set(cos(alfa),sin(alfa),sin(beta)); // vzorka
  v.normalize();                        // normalizacia
  v=v*(16.0f-i)/16.0f;                  // postupne zmensovanie velkosti
  v=v*radius;                           // polomer
  ptr[j+0]=v.x; ptr[j+1]=v.y; ptr[j+2]=v.z; // ulozenie hodnot do pola
  // INFO: printf("%f %f %f\n",ptr[j+0],ptr[j+1],ptr[j+2]);
  j+=3;

  alfa+=PI_DIV_2; // pre vypocet hodnot nasledujucej vzorky
  if (((i+1)%4)==0) { alfa+=PI_DIV_4; beta+=PI_DIV_2/4.0f; }
 }

 // scale-bias matica a inverzna scale-bias matica
 sbmat.set_scalebias(); isbmat.set_scalebias_inverse();
 // PROJ = scale-bias * PROJECTION
 sbproj=sbmat*avatar.get_pr_mat(0);
 // IPROJ = IPROJECTION * iscale-bias
 isbproj=mm.ipm(avatar.get_pr_mat(0))*isbmat;
 // light projection matica
 lproj.perspective(60.0f,(GLfloat) SHADOWMAPSIZE/(GLfloat) SHADOWMAPSIZE,0.1f,1000.0f);
 lsbproj=sbmat*lproj;

 ssao_shader[1].on();
 ssao_shader[1].bindattriblocation(0,"in_pos"); // vrcholy
 ssao_shader[1].uniformI("depth_buf_unit",0);   // texturovacia jednotka pre hlbky
 ssao_shader[1].uniformI("normal_buf_unit",1);  // texturovacia jednotka pre normaly
 ssao_shader[1].uniformI("rndvec_buf_unit",2);  // texturovacia jednotka pre nahodne orientovane vektory
 ssao_shader[1].uniformI("base_tex_unit",3);    // texturovacia jednotka pre farby
 ssao_shader[1].uniformFV("samples",16,ssao_samples); // vzorky
 ssao_shader[1].uniformmatrixFV("sbproj",1,sbproj.get_data());
 ssao_shader[1].uniformmatrixFV("isbproj",1,isbproj.get_data());
 ssao_shader[1].off();

 float offsetsh[6],offsetsv[6];
 loopi(5) {
  offsetsh[i]=(i-2.0f)/((float) scr_w/2.0f);
  offsetsv[i]=(i-2.0f)/((float) scr_h/2.0f);
 }
 offsetsh[5]=0.0f; offsetsv[5]=0.0f; // nepouzije sa
 float weights[6]={0.121786f,0.140999f,0.148054f,0.140999f,0.121786f,0.0f};

 ssao_shader[2].on();
 ssao_shader[2].bindattriblocation(0,"in_pos"); // vrcholy
 ssao_shader[2].bindattriblocation(8,"in_uv");  // texturovacie suradnice
 ssao_shader[2].uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 ssao_shader[2].uniformFV("offsetsh",2,offsetsh);
 ssao_shader[2].uniformFV("weights",2,weights);
 ssao_shader[2].off();

 ssao_shader[3].on();
 ssao_shader[3].bindattriblocation(0,"in_pos"); // vrcholy
 ssao_shader[3].bindattriblocation(8,"in_uv");  // texturovacie suradnice
 ssao_shader[3].uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 ssao_shader[3].uniformFV("offsetsv",2,offsetsv);
 ssao_shader[3].uniformFV("weights",2,weights);
 ssao_shader[3].off();

 ssao_shader[4].on();
 ssao_shader[4].bindattriblocation(0,"in_pos"); // vrcholy
 ssao_shader[4].bindattriblocation(8,"in_uv");  // texturovacie suradnice
 ssao_shader[4].uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 ssao_shader[4].off();

 shadow_shader.on();
 shadow_shader.bindattriblocation(0,"in_pos"); // vrcholy
 shadow_shader.bindattriblocation(4,"in_nor"); // normaly
 shadow_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 shadow_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 shadow_shader.uniformI("sm_tex_unit",1);      // shadowmap texturovacia jednotka
 shadow_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 shadow_shader.off();

 // pre deferred shading
 null_shader.on();
 null_shader.bindattriblocation(0,"in_pos"); // vrcholy
 null_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 null_shader.off();
 raw_shader.on();
 raw_shader.bindattriblocation(0,"in_pos"); // vrcholy
 raw_shader.bindattriblocation(2,"in_tan"); // tangenty
 raw_shader.bindattriblocation(3,"in_bit"); // bitangenty
 raw_shader.bindattriblocation(4,"in_nor"); // normaly
 raw_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 raw_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 raw_shader.uniformI("nm_tex_unit",1);      // normalmap texturovacia jednotka
 raw_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 raw_shader.off();
 def_shader.on();
 def_shader.bindattriblocation(0,"in_pos"); // vrcholy
 def_shader.uniformI("depth_tex_unit",0);   // texturovacia jednotka pre hlbky
 def_shader.uniformI("albedo_tex_unit",1);  // texturovacia jednotka pre albedo
 def_shader.uniformI("normal_tex_unit",2);  // texturovacia jednotka pre normaly
 def_shader.uniformmatrixFV("sbproj",1,sbproj.get_data());
 def_shader.uniformmatrixFV("isbproj",1,isbproj.get_data());
 // naplnenie pola difuznymi komponentami 16-tich svetelnych zdrojov
 a=0;
 loopi(16) {
  lightpos[a+0]=mlight[i].diffuse.x; lightpos[a+1]=mlight[i].diffuse.y;
  lightpos[a+2]=mlight[i].diffuse.z; lightpos[a+3]=mlight[i].diffuse.w;
  a+=4;
 }
 def_shader.uniformmatrixFV("light_dif",4,lightpos);
 def_shader.off();
 msaa_shader.on();
 msaa_shader.bindattriblocation(0,"in_pos"); // vrcholy
 msaa_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 msaa_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 msaa_shader.uniformI("msaa_smp",msaa_samples);                  // MSAA samples
 msaa_shader.uniform2F("dxdy",1.0f/(float) (scr_w*msaa_samples), // delta dvoch susediacich texturovacich suradnic
                              1.0f/(float) (scr_h*msaa_samples));
 msaa_shader.off();

 deltas_shader.on();
 deltas_shader.bindattriblocation(0,"in_pos"); // vrcholy
 deltas_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 deltas_shader.off();

 mblur_shader.on();
 mblur_shader.bindattriblocation(0,"in_pos"); // vrcholy
 mblur_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 mblur_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 mblur_shader.uniformI("deltas_tex_unit",1);  // texturovacia jednotka s pohybovymi vektormi
 mblur_shader.uniformI("msaa_smp",msaa_samples);                   // MSAA samples
 mblur_shader.uniform2F("dxdy",1.0f/(float) (scr_w*msaa_samples), // delta dvoch susediacich texturovacich suradnic
                               1.0f/(float) (scr_h*msaa_samples));
 mblur_shader.off();

 triangle_shrink.on();
 triangle_shrink.bindattriblocation(0,"in_pos"); // vrcholy
 triangle_shrink.bindattriblocation(1,"in_col"); // farby
 triangle_shrink.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 triangle_shrink.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 triangle_shrink.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 triangle_shrink.off();

 silhouette_shader.on();
 silhouette_shader.bindattriblocation(0,"in_pos"); // vrcholy
 silhouette_shader.bindattriblocation(1,"in_col"); // farby
 silhouette_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
 silhouette_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
 silhouette_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
 silhouette_shader.off();

 int g;
 glGetIntegerv(GL_MAX_DRAW_BUFFERS,&g); printf("GL_MAX_DRAW_BUFFERS: %i\n",g);

 lA=0.0f; lB=0.0f; // natocenie svetelneho zdroja

 state=0; // pomocny parameter (specificky stav) aplikacie

 return(0);
}

// nastartovanie casovaca **************************************************
// 1s=1000ms
// Uint32 SDL_GetTicks() vracia pocet ms od startu SDL,
// po 49.7 dnoch dojde k prekroceniu 32bit limitu a pocitadlo sa vynuluje
void CInterface::resettimer(void)
{
 // aktualny casovy okamih (ms)
 actTime=(GLuint) SDL_GetTicks();
 // povodny casovy okamih (ms)
 oldTime=actTime;
 // pocet uplynutych milisekund
 millis=actTime-oldTime;
 // pomocne premenne
 fpsdelay=0; fpserror=0;

 // uplynuty cas medzi aktualnym a povodnym casovym okamihom (s)
 dTime=((GLfloat) millis)/1000.0f;

 // uplynute ms
 mspos=0; mstotal=0;
 // pocet snimkov za sekundu, minimalny a maximalny pocet snimkov za sekundu
 actfps=0; minfps=0; maxfps=0;
}

// aktualny pocet fps, minimalny a maximalny pocet fps *********************
void CInterface::getfps(GLuint &ACT,GLuint &MIN,GLuint &MAX)
{
 // posledny prvok pola uplynutych ms
 mstotal=mshistory[MSHISTORY-1]; MIN=mstotal; MAX=mstotal;
 // ostatne prvky pola
 loopi(MSHISTORY-1) {
  mstotal+=mshistory[i];
  if (mshistory[i]>MIN) MIN=mshistory[i];
  if (mshistory[i]<MAX) MAX=mshistory[i];
 }
 ACT=(1000*MSHISTORY)/mstotal; MIN=1000/MIN; MAX=1000/MAX;
}

// aktualizacia casovaca ***************************************************
void CInterface::updatetimer(void)
{
 // aktualny casovy okamih (ms)
 actTime=(GLuint) SDL_GetTicks();
 // aktualny casovy okamih nemoze byt "starsi" ako povodny casovy okamih
 if (actTime<oldTime) actTime=oldTime;
 // pocet uplynutych milisekund
 millis=actTime-oldTime;
 // uprava poctu uplynutych milisekund
 fpsdelay=1000/MAXFPS-millis;
 if (fpsdelay<=0) fpserror=0; // ak MAXFPS=200 => ak millis>=5
 else {
  fpserror+=1000%MAXFPS; // ak MAXFPS=200 => fpserror+=0
  if (fpserror>=MAXFPS) { fpserror-=MAXFPS; fpsdelay++; }
  if (fpsdelay>0) { // ak millis<5
   SDL_Delay(fpsdelay); // pauza
   actTime+=(GLuint) fpsdelay; // aktualny casovy okamih (ms)
   millis=actTime-oldTime; // pocet uplynutych milisekund
  }
 }

 // uplynuty cas medzi aktualnym a povodnym casovym okamihom (s)
 dTime=((GLfloat) millis)/1000.0f;
 oldTime=actTime;

 // aktualizacia pola uplynutych ms
 mshistory[mspos++]=(GLuint) max(1,min(1000,(int) millis));
 // po naplneni pola uplynutych ms sa vypocita actfps, minfps a maxfps
 if (mspos>=MSHISTORY) {
  mspos=0;
  // aktualny pocet fps, minimalny a maximalny pocet fps
  getfps(actfps,minfps,maxfps);
  // INFO: fps
  //printf("%u %u %u\n",actfps,minfps,maxfps);
 }
}

// casovo zavisla uprava objektov, resp. sceny *****************************
void CInterface::OGL_idle(void)
{

}

// aktualizacia paramterov osvetlenia **************************************
void CInterface::update_lighting_params(void)
{
 glLightfv(GL_LIGHT0,GL_AMBIENT,lsrcparam.ambient.get_data());
 glLightfv(GL_LIGHT0,GL_DIFFUSE,lsrcparam.diffuse.get_data());
 glLightfv(GL_LIGHT0,GL_SPECULAR,lsrcparam.specular.get_data());
 glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,lsrcparam.spotDirection.get_data());
 glLightf(GL_LIGHT0,GL_SPOT_EXPONENT,lsrcparam.spotExponent);
 glLightf(GL_LIGHT0,GL_SPOT_CUTOFF,lsrcparam.spotCutoff);
 lsrcparam.spotCosCutoff=cos(DEG2RAD(lsrcparam.spotCutoff));
 lsrcparam.spotSinCutoff=sin(DEG2RAD(lsrcparam.spotCutoff));
 glLightf(GL_LIGHT0,GL_CONSTANT_ATTENUATION,lsrcparam.constantAttenuation);
 glLightf(GL_LIGHT0,GL_LINEAR_ATTENUATION,lsrcparam.linearAttenuation);
 glLightf(GL_LIGHT0,GL_QUADRATIC_ATTENUATION,lsrcparam.quadraticAttenuation);

 glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lmodelparam.ambient.get_data());

 glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,mparam.ambient.get_data());
 glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,mparam.diffuse.get_data());
 glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,mparam.specular.get_data());
 glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,mparam.emission.get_data());
 glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,mparam.shininess);
}

// vykreslenie OpenGL sceny ************************************************
void CInterface::OGL_draw(void)
{
 // vymazanie zasobnika farieb, hlbky a Stencil zasobnika
 //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
 // v pripade niektorych shaderov nie je potrebne
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

 // modelview matica bude upravena podla transformacnej matice Avatara
 mvmat=avatar.get_mv_mat();
 // uprava modelview matice
 glLoadMatrixf(mvmat.get_data());
 // umiestnenie svetelneho zdroja na danu poziciu
 // OpenGL polohu automaticky transformuje z OBJECT do EYE space
 glLightfv(GL_LIGHT0,GL_POSITION,lsrcparam.position.get_data());
 glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,lsrcparam.spotDirection.get_data());

 switch (but_text_id) {
  case 0: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          base_shader.on();
          // modelview matica
          base_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[1]]);
          objects_draw();
          base_shader.off();
          break;
  case 1: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          light_shader.on();
          // modelview matica
          light_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[1]]);
          objects_draw();
          light_shader.off();
          break;
  case 2: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          brick_shader.on();
          // modelview matica
          brick_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[1]]);
          objects_draw();
          brick_shader.off();
          break;
  case 3: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          toon_shader.on();
          // modelview matica
          toon_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[1]]);
          objects_draw();
          toon_shader.off();
          break;
  case 4: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          wave_shader.on();
          // aktualny casovy okamih
          wave_shader.uniform1F("at",actTime);
          // modelview matica
          wave_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[1]]);
          objects_draw();
          wave_shader.off();
          break;
  case 5: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          wood_shader.on();
          // modelview matica
          wood_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_3D,texture.tex_id[tid[3]]);
          objects_draw();
          wood_shader.off();
          break;
  case 6: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          nm_shader.on();
          // modelview matica
          nm_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textur
          glActiveTexture(GL_TEXTURE0);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[11]]);
          glActiveTexture(GL_TEXTURE1);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[12]]);
          objects_draw();
          glActiveTexture(GL_TEXTURE0);
          nm_shader.off();
          break;
  case 7: // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          px_shader.on();
          // modelview matica
          px_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
          // pripojenie textur
          glActiveTexture(GL_TEXTURE0);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[13]]);
          glActiveTexture(GL_TEXTURE1);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[14]]);
          glActiveTexture(GL_TEXTURE2);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[15]]);
          objects_draw();
          glActiveTexture(GL_TEXTURE0);
          px_shader.off();
          break;
  case 8: // priprava reflection textury
          draw_water_scene(0);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_reflection]);
          glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,512,512);

          // priprava refraction textury
          glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
          draw_water_scene(1);
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_refraction]);
          glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,512,512);
          // priprava depth textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_depth]);
          glCopyTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT,0,0,512,512,0);

          // vykreslenie kompletnej sceny s vodou
          glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
          draw_water_scene(2);
          break;
  case 9: // 1. krok - vykreslenie sceny do FBO1 - depth, texture
          glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo.getfbo1());
          glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
          glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

          dof_shader.on(0); // shader pre vykreslenie sceny a vypocet blur
          dof_shader.uniformmatrixFV(0,"modelview_mat",avatar.get_mv_mat());
          // pripojenie textury
          glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[16]]);
          // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
          objects_draw();
          dof_shader.off();

          // ulozenie potrebnych atributov
          glPushAttrib(GL_DEPTH_BUFFER_BIT);
          glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

          glMatrixMode(GL_PROJECTION); // projekcna matica
          glPushMatrix();              // ulozenie
          glLoadIdentity();            // reset

          // nastavenie pravouhleho premietania - vymeni Y
          glOrtho(0.0f,(GLfloat) scr_w,(GLfloat) scr_h,0.0f,-1.0f,1.0f);

          glMatrixMode(GL_MODELVIEW); // modelview matica
          glPushMatrix();             // ulozenie
          glLoadIdentity();           // reset

          glViewport(0,0,scr_w2,scr_h2); // polovicne rozmery okna

          // 2. krok - vykreslenie obsahu FBO1 do FBO2
          glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo.getfbo2());

          glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT); // kresli do FBO2 - texture0
          glClear(GL_COLOR_BUFFER_BIT);

          dof_shader.on(1); // shader pre jednoduche vykreslenie stvoruholnika
          fbo.drawfbo1_attachment(); // vykresli obsah FBO1 - texture
          dof_shader.off();

          // 3. krok - Gausov filter v smere X
          glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT); // kresli do FBO2 - texture1
          glClear(GL_COLOR_BUFFER_BIT);

          dof_shader.on(2); // aplikacia Gausovho filtra v smere X
          fbo.drawfbo2_attachment(0); // vykresli obsah FBO2 - texture0
          dof_shader.off();

          // 4. krok - Gausov filter v smere Y
          glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT); // kresli do FBO2 - texture2
          glClear(GL_COLOR_BUFFER_BIT);

          dof_shader.on(3); // aplikacia Gausovho filtra v smere Y
          fbo.drawfbo2_attachment(1); // vykresli obsah FBO2 - texture1
          dof_shader.off();

          // 5. krok - vysledne vykreslenie
          glViewport(0,0,scr_w,scr_h); // plne rozmery okna

          // vykreslenie sceny do aplikaciou alokovaneho buffra
          glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
          glDrawBuffer(GL_BACK);
          glActiveTexture(GL_TEXTURE0); // originalna scena
          glBindTexture(GL_TEXTURE_2D,fbo.getfbo1_tex_id()); // obsah FBO1 - texture
          glActiveTexture(GL_TEXTURE1); // Gausov filter
          glBindTexture(GL_TEXTURE_2D,fbo.getfbo2_tex_id(2)); // obsah FBO2 - texture2
          dof_shader.on(4); // zmiesanie originalnej sceny s Gaussian blur
          fbo.draw(); // vykresli stvoruholnik do celeho okna
          glActiveTexture(GL_TEXTURE0);
          dof_shader.off();

          glPopMatrix(); // obnova modelview matice
          glMatrixMode(GL_PROJECTION);
          glPopMatrix(); // obnova projection matice

          // aktualna je vzdy modelview matica
          glMatrixMode(GL_MODELVIEW);
          glPopAttrib(); // DEPTH_BUFFER_BIT

          break;
  case 10: // 1. krok - vykreslenie sceny do FBO0 - depth, texture
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,hdr_fbo.getfbo0());
           glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
           glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

           // vykreslenie okoliteho prostredia
           dof_shader.on(0); // shader tu neplni funkciu DOF shadera, je to iba pomocka pre vykreslenie sceny
           // modelview matica
           dof_shader.uniformmatrixFV(0,"modelview_mat",avatar.get_mv_mat());
           enviro.draw();
           dof_shader.off();

           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           chromatic_shader.on();
           // modelview matica
           chromatic_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // inverzna modelview matica
           chromatic_shader.uniformmatrixFV("inv_modelview_mat",1,avatar.get_inv_mv_mat());
           // pripojenie textur
           glEnable(GL_TEXTURE_CUBE_MAP);
           // prvych 6 textur je urcenych pre pokrytie vnutornych stien kocky
           // predstavujucej okolite prostredie, 7. textura je samotna CubeMap textura
           glBindTexture(GL_TEXTURE_CUBE_MAP,texture.tex_id[tid[20+6]]);
           objects_draw();
           glDisable(GL_TEXTURE_CUBE_MAP);
           chromatic_shader.off();

           // svietidlo - nutne vykreslit tu aby bolo zahrnute do bloom efektu
           glTranslatef(lsrcparam.position.x,lsrcparam.position.y,lsrcparam.position.z);
           glRotatef(lA,0.0f,1.0f,0.0f);
           glRotatef(lB,1.0f,0.0f,0.0f);
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[10]]);
           glBegin(GL_QUADS);
            glTexCoord2f(0.12f,0.055f); glVertex3f(-0.3f,-0.3f, 0.6f); glTexCoord2f(0.35f,0.055f); glVertex3f( 0.3f,-0.3f, 0.6f);
            glTexCoord2f(0.35f,0.282f); glVertex3f( 0.3f, 0.3f, 0.6f); glTexCoord2f(0.12f,0.282f); glVertex3f(-0.3f, 0.3f, 0.6f);

            glTexCoord2f(0.12f,0.0f);  glVertex3f(-0.3f,-0.3f, 0.6f); glTexCoord2f(0.35f,0.0f);  glVertex3f(-0.3f, 0.3f, 0.6f);
            glTexCoord2f(0.35f,0.35f); glVertex3f(-0.3f, 0.3f,-0.6f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.3f,-0.3f,-0.6f);
            glTexCoord2f(0.12f,0.0f);  glVertex3f( 0.3f, 0.3f, 0.6f); glTexCoord2f(0.35f,0.0f);  glVertex3f( 0.3f,-0.3f, 0.6f);
            glTexCoord2f(0.35f,0.35f); glVertex3f( 0.3f,-0.3f,-0.6f); glTexCoord2f(0.12f,0.35f); glVertex3f( 0.3f, 0.3f,-0.6f);
            glTexCoord2f(0.12f,0.0f);  glVertex3f(-0.3f, 0.3f, 0.6f); glTexCoord2f(0.35f,0.0f);  glVertex3f( 0.3f, 0.3f, 0.6f);
            glTexCoord2f(0.35f,0.35f); glVertex3f( 0.3f, 0.3f,-0.6f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.3f, 0.3f,-0.6f);
            glTexCoord2f(0.12f,0.0f);  glVertex3f( 0.3f,-0.3f, 0.6f); glTexCoord2f(0.35f,0.0f);  glVertex3f(-0.3f,-0.3f, 0.6f);
            glTexCoord2f(0.35f,0.35f); glVertex3f(-0.3f,-0.3f,-0.6f); glTexCoord2f(0.12f,0.35f); glVertex3f( 0.3f,-0.3f,-0.6f);

            glTexCoord2f(0.12f,0.35f); glVertex3f(-0.3f,  0.3f, -0.6f); glTexCoord2f(0.35f,0.35f); glVertex3f( 0.3f,  0.3f, -0.6f);
            glTexCoord2f(0.48f,0.5f);  glVertex3f( 0.45f, 0.45f,-1.2f); glTexCoord2f(0.0f,0.5f);   glVertex3f(-0.45f, 0.45f,-1.2f);
            glTexCoord2f(0.12f,0.35f); glVertex3f( 0.3f, -0.3f, -0.6f); glTexCoord2f(0.35f,0.35f); glVertex3f(-0.3f, -0.3f, -0.6f);
            glTexCoord2f(0.48f,0.5f);  glVertex3f(-0.45f,-0.45f,-1.2f); glTexCoord2f(0.0f,0.5f);   glVertex3f( 0.45f,-0.45f,-1.2f);
            glTexCoord2f(0.35f,0.35f); glVertex3f(-0.3f,  0.3f, -0.6f); glTexCoord2f(0.48f,0.5f);  glVertex3f(-0.45f, 0.45f,-1.2f);
            glTexCoord2f(0.0f,0.5f);   glVertex3f(-0.45f,-0.45f,-1.2f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.3f, -0.3f, -0.6f);
            glTexCoord2f(0.12f,0.35f); glVertex3f( 0.3f,  0.3f, -0.6f); glTexCoord2f(0.35f,0.35f); glVertex3f( 0.3f, -0.3f, -0.6f);
            glTexCoord2f(0.48f,0.5f);  glVertex3f( 0.45f,-0.45f,-1.2f); glTexCoord2f(0.0f,0.5f);   glVertex3f( 0.45f, 0.45f,-1.2f);

            glTexCoord2f(0.5f,0.5f); glVertex3f(-0.45f, 0.45f,-1.2f); glTexCoord2f(1.0f,0.5f); glVertex3f( 0.45f, 0.45f,-1.2f);
            glTexCoord2f(1.0f,1.0f); glVertex3f( 0.45f,-0.45f,-1.2f); glTexCoord2f(0.5f,1.0f); glVertex3f(-0.45f,-0.45f,-1.2f);
           glEnd();

           // ulozenie potrebnych atributov
           glPushAttrib(GL_DEPTH_BUFFER_BIT);
           glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

           glMatrixMode(GL_PROJECTION); // projekcna matica
           glPushMatrix();              // ulozenie
           glLoadIdentity();            // reset

           // nastavenie pravouhleho premietania - vymeni Y
           glOrtho(0.0f,(GLfloat) scr_w,(GLfloat) scr_h,0.0f,-1.0f,1.0f);

           glMatrixMode(GL_MODELVIEW); // modelview matica
           glPushMatrix();             // ulozenie
           glLoadIdentity();           // reset

           glViewport(0,0,scr_w,scr_h); // plne rozmery okna

           // 2. krok - vyber svetlych casti
           glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT); // kresli do FBO0 - texture1
           glClear(GL_COLOR_BUFFER_BIT);

           glBindTexture(GL_TEXTURE_2D,hdr_fbo.getfbo0_tex_id(0));
           bloom_shader.on();
           hdr_fbo.draw(); // vykresli obsah FBO0 - texture0
           bloom_shader.off();

           // 3. krok - downsampling
           blur_shader.on();
           a=scr_w,b=scr_h;
           loopi(4) { // scena sa postupne zmensi na 1/2 1/4 1/8 1/16 - tinu
            a/=2; b/=2;
            glViewport(0,0,a,b); // polovicne rozmery okna

            // i-ty FBO pre downsampler
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,downsampler_fbo.getid(i));
            glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
            glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
            if (i==0) glBindTexture(GL_TEXTURE_2D,hdr_fbo.getfbo0_tex_id(1)); // FBO0 - texture1
            else glBindTexture(GL_TEXTURE_2D,downsampler_fbo.gettex_id(i-1)); // obsah predosleho FBO

            vec3 v(a,b,0.0f);
            blur_shader.uniformFV("scr",1,v.get_data());
            downsampler_fbo.draw(); // vykresli obsah predosleho FBO do noveho FBO
           }
           blur_shader.off();

           glViewport(0,0,scr_w,scr_h); // plne rozmery okna

           // 4. krok - aditivne zmiesanie 4 FBO pre downsampler
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,hdr_fbo.getfbo0()); // kresli do FBO0 - texture1
           glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
           glClear(GL_COLOR_BUFFER_BIT);
           glEnable(GL_BLEND);
           glBlendFunc(GL_ONE,GL_ONE);
           dof_shader.on(1); // shader tu neplni funkciu DOF shadera, je to iba pomocka pre vykreslenie sceny
           loopi(4) {
            glBindTexture(GL_TEXTURE_2D,downsampler_fbo.gettex_id(i)); // obsah i-teho FBO pre downsampler
            downsampler_fbo.draw();
           }
           dof_shader.off();
           glDisable(GL_BLEND);

           // vykreslenie sceny do aplikaciou alokovaneho buffra
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
           glDrawBuffer(GL_BACK);
           glActiveTexture(GL_TEXTURE0); // originalna scena
           glBindTexture(GL_TEXTURE_2D,hdr_fbo.getfbo0_tex_id(0)); // obsah FBO0 - texture0
           glActiveTexture(GL_TEXTURE1); // bloom
           glBindTexture(GL_TEXTURE_2D,hdr_fbo.getfbo0_tex_id(1)); // obsah FBO0 - texture1
           //glBindTexture(GL_TEXTURE_2D,downsampler_fbo.gettex_id(state)); // obsah i-teho FBO pre downsampler
           hdr_shader.on();
           hdr_fbo.draw(); // vykresli stvoruholnik do celeho okna
           hdr_shader.off();
           glActiveTexture(GL_TEXTURE0);

           glPopMatrix(); // obnova modelview matice
           glMatrixMode(GL_PROJECTION);
           glPopMatrix(); // obnova projection matice

           // aktualna je vzdy modelview matica
           glMatrixMode(GL_MODELVIEW);
           glPopAttrib(); // DEPTH_BUFFER_BIT

           break;
  case 11: // 1. krok - vykreslenie sceny do FBO pre SSAO - depth, normal, color
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,ssao_fbo.getfbo_ssao());
           glDrawBuffers(2,clrsbuf);
           glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

           ssao_shader[0].on(); // naplnenie FBO pre SSAO hlbkou, normalami a farbami
           ssao_shader[0].uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // textura objektov
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[17]]);
           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           objects_draw();
           ssao_shader[0].off();

           // 2. krok - aplikacia SSAO, pri vykresleni nam pomoze FBO1,
           // ktory uz mame pripraveny a pouzivame ho aj pre ine shadery

           // ulozenie potrebnych atributov
           glPushAttrib(GL_DEPTH_BUFFER_BIT);
           glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

           glMatrixMode(GL_PROJECTION); // projekcna matica
           glPushMatrix();              // ulozenie
           glLoadIdentity();            // reset

           // nastavenie pravouhleho premietania - vymeni Y
           glOrtho(0.0f,(GLfloat) scr_w,(GLfloat) scr_h,0.0f,-1.0f,1.0f);

           glMatrixMode(GL_MODELVIEW); // modelview matica
           glPushMatrix();             // ulozenie
           glLoadIdentity();           // reset

           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo.getfbo1());
           glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
           glClear(GL_COLOR_BUFFER_BIT);
           glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,ssao_fbo.ssao_depth());  // hlbka
           glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,ssao_fbo.ssao_normal()); // normaly
           // 2D textura nahodne orientovanych vektorov
           glActiveTexture(GL_TEXTURE2); glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[4]]);
           glActiveTexture(GL_TEXTURE3); glBindTexture(GL_TEXTURE_2D,ssao_fbo.ssao_color());  // farby

           ssao_shader[1].on();
           ssao_fbo.draw(); // vykresli stvoruholnik do celeho okna
           ssao_shader[1].off();

           glActiveTexture(GL_TEXTURE0);
/*
           // 3. krok - H filter
           glBindTexture(GL_TEXTURE_2D,fbo.getfbo1_tex_id()); // obsah FBO1
           ssao_shader[2].on();
           ssao_fbo.draw(); // vykresli stvoruholnik do celeho okna
           ssao_shader[2].off();

           // 4. krok - V filter
           glBindTexture(GL_TEXTURE_2D,fbo.getfbo1_tex_id()); // obsah FBO1
           ssao_shader[3].on();
           ssao_fbo.draw(); // vykresli stvoruholnik do celeho okna
           ssao_shader[3].off();
*/
           // 5. krok - vysledne vykreslenie
           // vykreslenie sceny do aplikaciou alokovaneho buffra
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
           glDrawBuffer(GL_BACK);
           glBindTexture(GL_TEXTURE_2D,fbo.getfbo1_tex_id()); // obsah FBO1
           ssao_shader[4].on();
           ssao_fbo.draw(); // vykresli stvoruholnik do celeho okna
           ssao_shader[4].off();

           glPopMatrix(); // obnova modelview matice
           glMatrixMode(GL_PROJECTION);
           glPopMatrix(); // obnova projection matice

           // aktualna je vzdy modelview matica
           glMatrixMode(GL_MODELVIEW);
           glPopAttrib(); // DEPTH_BUFFER_BIT

           break;
  case 12: // 1. krok - priprava shadowmap - naplnenie textur informaciami o hlbke fragmentov
           draw_shadowmap();

           // 2. krok
           // textura objektov
           glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[18]]);
           // shadowmap
           glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D_ARRAY,shadowmap_fbo.shadowmap_depth());

           // kreslenie pomocou shadow shadera
           shadow_shader.on();
           shadow_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           shadow_shader.uniformmatrixFV("light_mat",5,lightmat); // matice svetelneho zdroja
           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           objects_draw();
           shadow_shader.off();

           // odpojenie textur
           glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D_ARRAY,0);
           glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,0);

           break;
  case 13: // 1. krok - vykreslenie sceny do FBO pre nespracovane udaje
           // msaa_samples-nasobne rozmery okna
           glViewport(0,0,msaa_samples*scr_w,msaa_samples*scr_h);

           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,raw_fbo.getfbo_raw());
           glDrawBuffers(3,clrsbuf);
           glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);

           // kreslenie pomocou raw shadera
           raw_shader.on();
           raw_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // pripojenie textur
           glActiveTexture(GL_TEXTURE0);
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[27]]);
           glActiveTexture(GL_TEXTURE1);
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[28]]);
           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           objects_draw();
           glActiveTexture(GL_TEXTURE0);
           raw_shader.off();

           // 2. krok - naplnenie Stencil zasobnika
           glDrawBuffer(GL_NONE); // bez vyplnania farieb
           glDepthMask(GL_FALSE); // bez vyplnania hlbiek

           glEnable(GL_STENCIL_TEST); // zapnutie Stencil testu
           // v zavislosti od testu s Depth zasobnikom sa hodnota v Stencil
           // zmeni vtedy, ak test hlbky skonci neuspechom
           // privratene plochy znizuju hodnotu v Stencil -1
           // odvratene plochy zvysuju hodnotu v Stencil +1
           glStencilFunc(GL_ALWAYS,0,0);
           glStencilOpSeparate(GL_BACK,GL_KEEP,GL_INCR,GL_KEEP);
           glStencilOpSeparate(GL_FRONT,GL_KEEP,GL_DECR,GL_KEEP);

           // nastavenie indexov na startovaciu hodnotu 0-15
           loopi(16) lind[i]=i;
           // zoradenie poloh svetelnych zdrojov podla Z-prvkov
           // prebieha v EYE space => je zavisle na polohe oka pozorovatela
           mm.quickSort(lightpos,lind,0,15);
           // spracuju sa aj odvratene plochy
           glDisable(GL_CULL_FACE);
           // kreslenie pomocou null shadera
           null_shader.on();

           glBindVertexArray(sphere_vao);
           loopi(16) {
            glPushMatrix();
            glTranslatef(mlight[lind[i]].position.x,mlight[lind[i]].position.y,mlight[lind[i]].position.z);
            glDrawArrays(GL_TRIANGLES,0,576/3);
            glPopMatrix();
           }
           glBindVertexArray(0);

           null_shader.off();
           glEnable(GL_CULL_FACE);

           glDepthMask(GL_TRUE); // povolenie vyplnania hlbiek

           // 3. krok - deferred shading
           // v dalsom budu spracovavane iba tie fragmenty, na pozicii ktorych
           // nie je v Stencil zasobniku hodnota 0
           glStencilFunc(GL_NOTEQUAL,0,0xFF);

           // ulozenie potrebnych atributov
           glPushAttrib(GL_DEPTH_BUFFER_BIT);
           glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

           glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT); // kresli do RAW -> GL_COLOR_ATTACHMENT0_EXT

           def_shader.on();
           // pripojenie textur - obsah FBO pre nespracovane udaje
           glActiveTexture(GL_TEXTURE0); // depth
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(3));
           glActiveTexture(GL_TEXTURE1); // albedo
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(1));
           glActiveTexture(GL_TEXTURE2); // normal
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(2));

           // naplnenie pola polohami 16-tich svetelnych zdrojov
           a=0;
           loopi(16) {
            // zmena polohy i-teho svetelneho zdroja
            if (state==0) {
             mlight[i].position.x=3.0f*sin(mlight[i].angle.x);
             mlight[i].position.y=3.0f*cos(mlight[i].angle.y)+5.0f;
             mlight[i].position.z=3.0f*sin(mlight[i].angle.z);
             mlight[i].angle+=mlight[i].omega;
            }
            // poloha i-teho svetelneho zdroja v EYE space
            lp=mvmat*mlight[i].position;
            // presun polohy do pola
            lightpos[a+0]=lp.x; lightpos[a+1]=lp.y; lightpos[a+2]=lp.z; lightpos[a+3]=lp.w;
            a+=4;
           }
           // 4 matice s polohami 16-tich svetelnych zdrojov
           def_shader.uniformmatrixFV("light_pos",4,lightpos);
           raw_fbo.draw(); // vykresli stvoruholnik do celeho okna
           glActiveTexture(GL_TEXTURE0);
           def_shader.off();

           glDisable(GL_STENCIL_TEST);

           // 4. krok - vysledne vykreslenie sceny
           // standardne rozmery okna
           glViewport(0,0,scr_w,scr_h);

           // vykreslenie sceny do aplikaciou alokovaneho buffra
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
           glDrawBuffer(GL_BACK);

           msaa_shader.on();
           // pripojenie textur - obsah RAW -> GL_COLOR_ATTACHMENT0_EXT
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(1));
           raw_fbo.draw(); // vykresli stvoruholnik do celeho okna
           msaa_shader.off();

           glPopAttrib(); // DEPTH_BUFFER_BIT

           break;
  case 14: // 1. krok - naplnenie prislusnej FBO textury pohybovymi vektormi
           // msaa_samples-nasobne rozmery okna
           glViewport(0,0,msaa_samples*scr_w,msaa_samples*scr_h);

           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,raw_fbo.getfbo_raw());
           glDrawBuffers(4,clrsbuf);
           glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);

           glDepthMask(GL_FALSE); // bez vyplnania hlbiek

           glDrawBuffer(GL_COLOR_ATTACHMENT3_EXT); // kresli do RAW -> GL_COLOR_ATTACHMENT3_EXT
           // kreslenie pomocou deltas shadera
           deltas_shader.on();
           deltas_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           deltas_shader.uniformmatrixFV("prev_modelview_mat",1,avatar.get_prev_mv_mat());
           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           objects_draw();
           deltas_shader.off();

           glDepthMask(GL_TRUE); // povolenie vyplnania hlbiek

           // 2. krok - vykreslenie sceny do FBO
           // kresli do RAW -> GL_COLOR_ATTACHMENT0_EXT,...1_EXT,...2_EXT
           glDrawBuffers(3,clrsbuf);
           // kreslenie pomocou raw shadera
           raw_shader.on();
           raw_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // pripojenie textur
           glActiveTexture(GL_TEXTURE0);
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[29]]);
           glActiveTexture(GL_TEXTURE1);
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[30]]);
           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           objects_draw();
           glActiveTexture(GL_TEXTURE0);
           raw_shader.off();

           // 3. krok - deferred shading
           glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT); // kresli do RAW -> GL_COLOR_ATTACHMENT0_EXT

           // nastavenie indexov na startovaciu hodnotu 0-15
           loopi(16) lind[i]=i;
           // zoradenie poloh svetelnych zdrojov podla Z-prvkov
           // prebieha v EYE space => je zavisle na polohe oka pozorovatela
           mm.quickSort(lightpos,lind,0,15);

           def_shader.on();
           // pripojenie textur - obsah FBO pre nespracovane udaje
           glActiveTexture(GL_TEXTURE0); // depth
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(3));
           glActiveTexture(GL_TEXTURE1); // albedo
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(1));
           glActiveTexture(GL_TEXTURE2); // normal
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(2));
           // naplnenie pola polohami 16-tich svetelnych zdrojov
           a=0;
           loopi(16) {
            // zmena polohy i-teho svetelneho zdroja
            if (state==0) {
             mlight[i].position.x=3.0f*sin(mlight[i].angle.x);
             mlight[i].position.y=3.0f*cos(mlight[i].angle.y)+5.0f;
             mlight[i].position.z=3.0f*sin(mlight[i].angle.z);
             mlight[i].angle+=mlight[i].omega;
            }
            // poloha i-teho svetelneho zdroja v EYE space
            lp=mvmat*mlight[i].position;
            // presun polohy do pola
            lightpos[a+0]=lp.x; lightpos[a+1]=lp.y; lightpos[a+2]=lp.z; lightpos[a+3]=lp.w;
            a+=4;
           }
           // 4 matice s polohami 16-tich svetelnych zdrojov
           def_shader.uniformmatrixFV("light_pos",4,lightpos);
           raw_fbo.draw(); // vykresli stvoruholnik do celeho okna
           glActiveTexture(GL_TEXTURE0);
           def_shader.off();

           // 4. krok - vysledne vykreslenie sceny s aplikaciou motion blur
           // standardne rozmery okna
           glViewport(0,0,scr_w,scr_h);

           // vykreslenie sceny do aplikaciou alokovaneho buffra
           glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
           glDrawBuffer(GL_BACK);

           // ulozenie potrebnych atributov
           glPushAttrib(GL_DEPTH_BUFFER_BIT);
           glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

           mblur_shader.on();
           mblur_shader.uniform1F("fps_correction",(float) actfps/15.0f);
           // pripojenie textur
           glActiveTexture(GL_TEXTURE0); // obsah RAW -> GL_COLOR_ATTACHMENT0_EXT
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(1));
           glActiveTexture(GL_TEXTURE1); // obsah RAW -> GL_COLOR_ATTACHMENT3_EXT
           glBindTexture(GL_TEXTURE_2D,raw_fbo.getfboraw_tex_id(4));
           raw_fbo.draw(); // vykresli stvoruholnik do celeho okna
           glActiveTexture(GL_TEXTURE0);
           mblur_shader.off();

           glPopAttrib(); // DEPTH_BUFFER_BIT

           break;
  case 15: glDisable(GL_CULL_FACE);

           // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
           triangle_shrink.on();
           // uroven zuzenia
           triangle_shrink.uniform1F("sl",shrink_level);
           // uhol rotacie
           triangle_shrink.uniform1F("ra",rotate_angle);
           rotate_angle+=PI_DIV_180*rotate_speed*2.0f;
           if (rotate_angle>PI_MUL_2) rotate_angle=0.0f;
           // modelview matica
           triangle_shrink.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // pripojenie textury
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[31]]);
           objects_draw();
           triangle_shrink.off();

           glEnable(GL_CULL_FACE);

           break;
  case 16: // vykreslenie objektov - kocka, gula, valec, ihlan
           silhouette_shader.on();
           // modelview matica
           silhouette_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
           // pripojenie textury
           glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[31]]);
           objects_draw();
           silhouette_shader.off();

           break;
 }

 glColor4f(1.0f,1.0f,1.0f,1.0f); // biela
/*
 // vykreslenie TBN
 glPushAttrib(GL_TEXTURE_BIT);
 glDisable(GL_TEXTURE_2D);

 glBegin(GL_LINES);
  for (int i=0;i<4824;i+=3) {
   glColor4f(1.0f,0.0f,0.0f,1.0f);
   glVertex3f(v_a[i+0],v_a[i+1],v_a[i+2]);
   glVertex3f(v_a[i+0]+tan_a[i+0]/8.0f,v_a[i+1]+tan_a[i+1]/8.0f,v_a[i+2]+tan_a[i+2]/8.0f);
   glColor4f(0.0f,1.0f,0.0f,1.0f);
   glVertex3f(v_a[i+0],v_a[i+1],v_a[i+2]);
   glVertex3f(v_a[i+0]+bit_a[i+0]/8.0f,v_a[i+1]+bit_a[i+1]/8.0f,v_a[i+2]+bit_a[i+2]/8.0f);
   glColor4f(0.0f,0.0f,1.0f,1.0f);
   glVertex3f(v_a[i+0],v_a[i+1],v_a[i+2]);
   glVertex3f(v_a[i+0]+nor_a[i+0]/2.0f,v_a[i+1]+nor_a[i+1]/2.0f,v_a[i+2]+nor_a[i+2]/2.0f);
  }
 glEnd();

 glColor4f(1.0f,1.0f,1.0f,1.0f); // biela

 glPopAttrib();
*/
/*
 // vykreslenie svetelneho zdroja
 // billboarding
 v[0].set(0.3f,0.0f,0.0f);
 s.x=(v[0].x*mvmat[0]);
 s.y=(v[0].x*mvmat[4]);
 s.z=(v[0].x*mvmat[8]);
 v[0].set(0.0f,0.3f,0.0f);
 t.x=(v[0].y*mvmat[1]);
 t.y=(v[0].y*mvmat[5]);
 t.z=(v[0].y*mvmat[9]);
 v[0].set(-s.x-t.x,-s.y-t.y,-s.z-t.z); // 1. vrchol
 v[1].set( s.x-t.x, s.y-t.y, s.z-t.z); // 2. vrchol
 v[2].set( s.x+t.x, s.y+t.y, s.z+t.z); // 3. vrchol
 v[3].set(-s.x+t.x,-s.y+t.y,-s.z+t.z); // 4. vrchol

 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[10]]);
 glTranslatef(lsrcparam.position.x,lsrcparam.position.y,lsrcparam.position.z);
 glBegin(GL_QUADS);
  glTexCoord2f(0.0f,0.0f); glVertex3fv(v[0].get_data());
  glTexCoord2f(1.0f,0.0f); glVertex3fv(v[1].get_data());
  glTexCoord2f(1.0f,1.0f); glVertex3fv(v[2].get_data());
  glTexCoord2f(0.0f,1.0f); glVertex3fv(v[3].get_data());
 glEnd();
 glDisable(GL_BLEND);
*/
 // vykresli sa iba ak nie je zvoleny HDR shader, alebo deferred shading
 if (but_text_id!=10 && but_text_id!=13 && but_text_id!=14) {
  glTranslatef(lsrcparam.position.x,lsrcparam.position.y,lsrcparam.position.z);
  glRotatef(lA,0.0f,1.0f,0.0f);
  glRotatef(lB,1.0f,0.0f,0.0f);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[10]]);
  glBegin(GL_QUADS);
   glTexCoord2f(0.12f,0.055f); glVertex3f(-0.1f,-0.1f, 0.2f); glTexCoord2f(0.35f,0.055f); glVertex3f( 0.1f,-0.1f, 0.2f);
   glTexCoord2f(0.35f,0.282f); glVertex3f( 0.1f, 0.1f, 0.2f); glTexCoord2f(0.12f,0.282f); glVertex3f(-0.1f, 0.1f, 0.2f);

   glTexCoord2f(0.12f,0.0f);  glVertex3f(-0.1f,-0.1f, 0.2f); glTexCoord2f(0.35f,0.0f);  glVertex3f(-0.1f, 0.1f, 0.2f);
   glTexCoord2f(0.35f,0.35f); glVertex3f(-0.1f, 0.1f,-0.2f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.1f,-0.1f,-0.2f);
   glTexCoord2f(0.12f,0.0f);  glVertex3f( 0.1f, 0.1f, 0.2f); glTexCoord2f(0.35f,0.0f);  glVertex3f( 0.1f,-0.1f, 0.2f);
   glTexCoord2f(0.35f,0.35f); glVertex3f( 0.1f,-0.1f,-0.2f); glTexCoord2f(0.12f,0.35f); glVertex3f( 0.1f, 0.1f,-0.2f);
   glTexCoord2f(0.12f,0.0f);  glVertex3f(-0.1f, 0.1f, 0.2f); glTexCoord2f(0.35f,0.0f);  glVertex3f( 0.1f, 0.1f, 0.2f);
   glTexCoord2f(0.35f,0.35f); glVertex3f( 0.1f, 0.1f,-0.2f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.1f, 0.1f,-0.2f);
   glTexCoord2f(0.12f,0.0f);  glVertex3f( 0.1f,-0.1f, 0.2f); glTexCoord2f(0.35f,0.0f);  glVertex3f(-0.1f,-0.1f, 0.2f);
   glTexCoord2f(0.35f,0.35f); glVertex3f(-0.1f,-0.1f,-0.2f); glTexCoord2f(0.12f,0.35f); glVertex3f( 0.1f,-0.1f,-0.2f);

   glTexCoord2f(0.12f,0.35f); glVertex3f(-0.1f,  0.1f, -0.2f); glTexCoord2f(0.35f,0.35f); glVertex3f( 0.1f,  0.1f, -0.2f);
   glTexCoord2f(0.48f,0.5f);  glVertex3f( 0.15f, 0.15f,-0.4f); glTexCoord2f(0.0f,0.5f);   glVertex3f(-0.15f, 0.15f,-0.4f);
   glTexCoord2f(0.12f,0.35f); glVertex3f( 0.1f, -0.1f, -0.2f); glTexCoord2f(0.35f,0.35f); glVertex3f(-0.1f, -0.1f, -0.2f);
   glTexCoord2f(0.48f,0.5f);  glVertex3f(-0.15f,-0.15f,-0.4f); glTexCoord2f(0.0f,0.5f);   glVertex3f( 0.15f,-0.15f,-0.4f);
   glTexCoord2f(0.35f,0.35f); glVertex3f(-0.1f,  0.1f, -0.2f); glTexCoord2f(0.48f,0.5f);  glVertex3f(-0.15f, 0.15f,-0.4f);
   glTexCoord2f(0.0f,0.5f);   glVertex3f(-0.15f,-0.15f,-0.4f); glTexCoord2f(0.12f,0.35f); glVertex3f(-0.1f, -0.1f, -0.2f);
   glTexCoord2f(0.12f,0.35f); glVertex3f( 0.1f,  0.1f, -0.2f); glTexCoord2f(0.35f,0.35f); glVertex3f( 0.1f, -0.1f, -0.2f);
   glTexCoord2f(0.48f,0.5f);  glVertex3f( 0.15f,-0.15f,-0.4f); glTexCoord2f(0.0f,0.5f);   glVertex3f( 0.15f, 0.15f,-0.4f);

   glTexCoord2f(0.5f,0.5f); glVertex3f(-0.15f, 0.15f,-0.4f); glTexCoord2f(1.0f,0.5f); glVertex3f( 0.15f, 0.15f,-0.4f);
   glTexCoord2f(1.0f,1.0f); glVertex3f( 0.15f,-0.15f,-0.4f); glTexCoord2f(0.5f,1.0f); glVertex3f(-0.15f,-0.15f,-0.4f);
  glEnd();
  glPushAttrib(GL_TEXTURE_BIT);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINES);
   glColor4f(0.5f,0.5f,0.5f,1.0f);
   glVertex3f(0.0f,0.0f,-0.4f); glVertex3f(0.0f,0.0f,-5.0f);
   glColor4f(1.0f,1.0f,1.0f,1.0f);
   glVertex3f(0.0f,0.0f,-0.4f); glVertex3f(5.0f*lsrcparam.spotSinCutoff,0.0f,-5.0f*lsrcparam.spotCosCutoff);
   glVertex3f(0.0f,0.0f,-0.4f); glVertex3f(-5.0f*lsrcparam.spotSinCutoff,0.0f,-5.0f*lsrcparam.spotCosCutoff);
   glVertex3f(0.0f,0.0f,-0.4f); glVertex3f(0.0f,5.0f*lsrcparam.spotSinCutoff,-5.0f*lsrcparam.spotCosCutoff);
   glVertex3f(0.0f,0.0f,-0.4f); glVertex3f(0.0f,-5.0f*lsrcparam.spotSinCutoff,-5.0f*lsrcparam.spotCosCutoff);
  glEnd();
  glPopAttrib();
 }
 if (but_text_id==13 || but_text_id==14) {
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[19]]);
  // billboarding castic predstavujucich svetelne zdroje
  v[0].set(0.2f,0.0f,0.0f);
  s.x=(v[0].x*mvmat[0]); s.y=(v[0].x*mvmat[4]); s.z=(v[0].x*mvmat[8]);
  v[0].set(0.0f,0.2f,0.0f);
  t.x=(v[0].y*mvmat[1]); t.y=(v[0].y*mvmat[5]); t.z=(v[0].y*mvmat[9]);
  v[0].set(-s.x-t.x,-s.y-t.y,-s.z-t.z); // 1. vrchol
  v[1].set( s.x-t.x, s.y-t.y, s.z-t.z); // 2. vrchol
  v[2].set( s.x+t.x, s.y+t.y, s.z+t.z); // 3. vrchol
  v[3].set(-s.x+t.x,-s.y+t.y,-s.z+t.z); // 4. vrchol
  loopi(16) {
   glColor4f(mlight[lind[i]].diffuse.x,mlight[lind[i]].diffuse.y,mlight[lind[i]].diffuse.z,1.0f);
   glPushMatrix();
   glTranslatef(mlight[lind[i]].position.x,mlight[lind[i]].position.y,mlight[lind[i]].position.z);
   glBegin(GL_QUADS);
    glTexCoord2f(0.0f,0.0f); glVertex3fv(v[0].get_data());
    glTexCoord2f(1.0f,0.0f); glVertex3fv(v[1].get_data());
    glTexCoord2f(1.0f,1.0f); glVertex3fv(v[2].get_data());
    glTexCoord2f(0.0f,1.0f); glVertex3fv(v[3].get_data());
   glEnd();
   glPopMatrix();
  }
  glColor4f(1.0f,1.0f,1.0f,1.0f);
  glDisable(GL_BLEND);
 }
 // vykreslenie head-up display
 drawhud();
}

// viacnasobne vykreslenie sceny s vodou ***********************************
void CInterface::draw_water_scene(int PASS)
{
 if (PASS<2) glViewport(0,0,512,512);
 else glViewport(0,0,(GLfloat) scr_w,(GLfloat) scr_h);

 if (PASS==0) { // reflection texture
  glCullFace(GL_FRONT);

  wtr.calc_water_plane(0.0f,-1.0f,0.0f,1.0f);
  glClipPlane(GL_CLIP_PLANE0,wtr.water_plane);
  glEnable(GL_CLIP_PLANE0);

  avatar.mv_mat_translate(0.0f,2.0f*wtr.water_level,0.0f);
  avatar.mv_mat_scale(1.0f,-1.0f,1.0f);
 }

 if (PASS==1) { // refraction texture
  wtr.calc_water_plane(0.0f,-1.0f,0.0f,1.0f);
  glClipPlane(GL_CLIP_PLANE0,wtr.water_plane);
  glEnable(GL_CLIP_PLANE0);
 }

 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 px_shader.on();
 // modelview matica
 px_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
 // pripojenie textur
 glActiveTexture(GL_TEXTURE0);
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[13]]);
 glActiveTexture(GL_TEXTURE1);
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[14]]);
 glActiveTexture(GL_TEXTURE2);
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[15]]);
 objects_draw();
 glActiveTexture(GL_TEXTURE0);
 px_shader.off();

 if (PASS==0) {
  avatar.mv_mat_scale(1.0f,-1.0f,1.0f);
  avatar.mv_mat_translate(0.0f,-2.0f*wtr.water_level,0.0f);

  glDisable(GL_CLIP_PLANE0);

  glCullFace(GL_BACK);
 }

 if (PASS==1) {
  glDisable(GL_CLIP_PLANE0);
 }

 if (PASS==2) { // kompletna scena s vodou
  // vykreslenie objektu predstavujuceho hladinu vody
  water_shader.on();
  // modelview matica
  water_shader.uniformmatrixFV("modelview_mat",1,avatar.get_mv_mat());
  // pripojenie textur
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_reflection]);
  glActiveTexture(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_refraction]);
  glActiveTexture(GL_TEXTURE2);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_depth]);
  glActiveTexture(GL_TEXTURE3);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_nm]);
  glActiveTexture(GL_TEXTURE4);
  glBindTexture(GL_TEXTURE_2D,texture.tex_id[wtr.tex_dudv]);
  // aktualny casovy okamih
  water_shader.uniform1F("at",actTime/10000.0f);

  wtr.draw();

  glActiveTexture(GL_TEXTURE0);
  water_shader.off();
 }
}

// vykreslenie sceny do shadowmap ******************************************
void CInterface::draw_shadowmap(void)
{
 // rozmery okna podla rozmerov shadowmap
 glViewport(0,0,SHADOWMAPSIZE,SHADOWMAPSIZE);

 // pripojenie FBO pre shadowmap - naplnaju sa iba hlbky, nie RGBA
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,shadowmap_fbo.getfbo_shadowmap());

 glCullFace(GL_FRONT);

 glMatrixMode(GL_PROJECTION);     // projekcna matica
 glPushMatrix();                  // ulozenie
 glLoadMatrixf(lproj.get_data()); // nastavenie podla lproj

 glMatrixMode(GL_MODELVIEW); // modelview matica
 glPushMatrix();             // ulozenie
 glLoadIdentity();           // jednotkova

 // natocenie a posun svetelenho zdroja = transformacia MODELVIEW matice
 glRotatef(-lB,1.0f,0.0f,0.0f);
 glRotatef(-lA,0.0f,1.0f,0.0f);
 glTranslatef(-lsrcparam.position.x,-lsrcparam.position.y,-lsrcparam.position.z);

 // presun prvkov MODELVIEW matice do matice mat
 glGetFloatv(GL_MODELVIEW_MATRIX,mat);

 // transformacna matica svetelneho zdroja
 lview.load(mat);
 // WC (world coords) = vertex coords
 // EC (eye coords) = modelview_matrix * WC
 // CC (clip coords) = projection_matrix * EC
 // NDC (normalized device coords) = CC.xyz / CC.w
 // WNDC (window coords) = scale-bias_matrix * NDC

 // matica svetelneho zdroja = scale-bias * projection * modelview
 // perspective divide je vykonavane v ramci FP
 lmat=lsbproj*lview;

 a=0;

 // 1. matica - svetelny zdroj v priamom smere
 loopi(16) lightmat[a+i]=lmat[i]; // 1. matica pre FP (fragment program)
 a+=16;
 // kresli do FBO pre shadowmap - 0. vrstva
 shadowmap_fbo.attach_layer(0);
 glClear(GL_DEPTH_BUFFER_BIT);
 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 objects_draw();

 // postupne natacanie sveteleneho zdroja okolo osi X a Y

 // 2. matica
 glPushMatrix();
 glRotatef(-1.0f,0.0f,1.0f,0.0f); // otocenie -1.0 okolo osi Y
 // presun prvkov MODELVIEW matice do matice mat
 glGetFloatv(GL_MODELVIEW_MATRIX,mat);
 lview.load(mat); lmat=lsbproj*lview;
 loopi(16) lightmat[a+i]=lmat[i]; // pridanie 2. matice pre FP
 a+=16;
 // kresli do FBO pre shadowmap - 1. vrstva
 shadowmap_fbo.attach_layer(1);
 glClear(GL_DEPTH_BUFFER_BIT);
 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 objects_draw();
 // 3. matica
 glRotatef(2.0f,0.0f,1.0f,0.0f); // otocenie +1.0 okolo osi Y
 // presun prvkov MODELVIEW matice do matice mat
 glGetFloatv(GL_MODELVIEW_MATRIX,mat);
 lview.load(mat); lmat=lsbproj*lview;
 loopi(16) lightmat[a+i]=lmat[i]; // pridanie 3. matice pre FP
 a+=16;
 // kresli do FBO pre shadowmap - 2. vrstva
 shadowmap_fbo.attach_layer(2);
 glClear(GL_DEPTH_BUFFER_BIT);
 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 objects_draw();
 glPopMatrix();

 // 4. matica
 glPushMatrix();
 glRotatef(-1.0f,1.0f,0.0f,0.0f); // otocenie -1.0 okolo osi X
 // presun prvkov MODELVIEW matice do matice mat
 glGetFloatv(GL_MODELVIEW_MATRIX,mat);
 lview.load(mat); lmat=lsbproj*lview;
 loopi(16) lightmat[a+i]=lmat[i]; // pridanie 4. matice pre FP
 a+=16;
 // kresli do FBO pre shadowmap - 3. vrstva
 shadowmap_fbo.attach_layer(3);
 glClear(GL_DEPTH_BUFFER_BIT);
 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 objects_draw();
 // 5. matica
 glRotatef(2.0f,1.0f,0.0f,0.0f); // otocenie +1.0 okolo osi X
 // presun prvkov MODELVIEW matice do matice mat
 glGetFloatv(GL_MODELVIEW_MATRIX,mat);
 lview.load(mat); lmat=lsbproj*lview;
 loopi(16) lightmat[a+i]=lmat[i]; // pridanie 5. matice pre FP
 // kresli do FBO pre shadowmap - 4. vrstva
 shadowmap_fbo.attach_layer(4);
 glClear(GL_DEPTH_BUFFER_BIT);
 // vykreslenie objektov - kocka, gula, valec, ihlan, plochy
 objects_draw();
 glPopMatrix();

 glPopMatrix(); // obnova modelview matice

 glMatrixMode(GL_PROJECTION);
 glPopMatrix(); // obnova projection matice

 // aktualna je vzdy modelview matica
 glMatrixMode(GL_MODELVIEW);

 glCullFace(GL_BACK);

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
 glDrawBuffer(GL_BACK);

 glViewport(0,0,scr_w,scr_h); // plne rozmery okna
}

// vykreslenie head-up display *********************************************
void CInterface::drawhud(void)
{
 // ulozenie potrebnych atributov
 glPushAttrib(GL_DEPTH_BUFFER_BIT);

 glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky
 glEnable(GL_BLEND);       // zapnutie priesvitnosti

 glMatrixMode(GL_PROJECTION); // projekcna matica
 glPushMatrix();              // ulozenie
 glLoadIdentity();            // reset

 // nastavenie pravouhleho premietania - vymeni Y
 glOrtho(0.0f,(GLfloat) scr_w,(GLfloat) scr_h,0.0f,-1.0f,1.0f);

 glMatrixMode(GL_MODELVIEW); // modelview matica
 glPushMatrix();             // ulozenie
 glLoadIdentity();           // reset

 // funkcia pre vypocet priesvitnosti
 glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);

 // vykreslenie menu -------------------------------------------------------

 // pripojenie prazdnej textury
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[0]]);
 // farba (priesvitnost) menu
 glColor4f(1.0f,1.0f,1.0f,0.1f);
 // stvoruholnik
 glBegin(GL_QUADS); // glOrtho() vymeni Y
  glVertex3f(5.0f,  5.0f,               0.0f);
  glVertex3f(5.0f,  (GLfloat) (scr_h-5),0.0f);
  glVertex3f(150.f, (GLfloat) (scr_h-5),0.0f);
  glVertex3f(150.0f,5.0f,               0.0f);
 glEnd();

 // vykreslenie tlacidiel
 loopi(BUTNUM) but[i].draw();

 // vypis textu ------------------------------------------------------------

 // pripojenie textury s fontom
 glBindTexture(GL_TEXTURE_2D,text.tex_id);

 // velkost textu
 text.scale(0.5f,0.5f);
 // farba textu
 glColor4f(0.8f,0.8f,0.0f,1.0f);
 // pocet fps
 text.drawf(15.0f,15.0f,"%u",actfps);
 // poloha Avatara
 text.drawf(15.0f,45.0f,"%.1f %.1f %.1f",-avatar.actpos.x,-avatar.actpos.y,-avatar.actpos.z);

 // velkost textu
 text.scale(0.4f,0.4f);
 if (but_text_id!=10 && but_text_id!=15) { // vykresli sa iba ak nie je zvoleny HDR shader, alebo triangle_shrink
  // parametre svetelneho zdroja
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,120.0f,"Light source");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,154.0,"ambient:");
  text.drawf(25.0f,229.0,"diffuse:");
  text.drawf(25.0f,304.0f,"specular:");
  text.drawf(25.0f,379.0f,"position:");
  text.drawf(25.0f,454.0f,"spotDirection:");
  text.drawf(25.0f,529.0f,"spotExponent:");
  text.drawf(25.0f,604.0f,"spotCutoff:");
  text.drawf(25.0f,679.0f,"constantAtt:");
  text.drawf(25.0f,754.0f,"linearAtt:");
  text.drawf(25.0f,829.0f,"quadraticAtt:");
  text.drawf(25.0f,904.0f,"global ambient:");
  // parametre materialu
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,995.0f,"Material");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,1029.0f,"ambient:");
  text.drawf(25.0f,1104.0f,"diffuse:");
  text.drawf(25.0f,1179.0f,"specular:");
  text.drawf(25.0f,1254.0f,"emission:");
  text.drawf(25.0f,1329.0f,"shininess:");
 }
 else if (but_text_id==10) { // pre HDR shader
  // refraction
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,120.0f,"Refraction");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,154.0,"red:");
  text.drawf(25.0f,229.0,"green:");
  text.drawf(25.0f,304.0f,"blue:");
  // fresnel
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,395.0f,"Fresnel");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,429.0f,"bias:");
  text.drawf(25.0f,504.0f,"scale:");
  text.drawf(25.0f,579.0f,"power:");
  // bloom
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,670.0f,"Bloom");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,704.0f,"factor:");
  text.drawf(25.0f,779.0f,"bright:");
  text.drawf(25.0f,854.0f,"exposure:");
 }
 else if (but_text_id==15) { // pre triangle_shrink
  // adjust geometry change
  glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,120.0f,"Adjust");
  glColor4f(0.8f,0.8f,0.0f,1.0f);
  text.drawf(25.0f,154.0,"shrink level:");
  text.drawf(25.0f,229.0,"rotation speed:");
 }

 // pouzity shader
 glColor4f(0.8f,0.8f,0.8f,1.0f); text.drawf(25.0f,1404.0f,"Shader");

 // vykreslenie zameriavacieho kriza ---------------------------------------

 // funkcia pre vypocet priesvitnosti
 glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

 // farba zameriavacieho kriza
 glColor4f(0.5f,0.5f,0.5f,1.0f);
 // pripojenie textury zameriavacieho kriza
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[2]]);

 glPushMatrix(); // ulozenie modelview matice
 // zameriavaci kriz na stred okna
 glTranslatef((GLfloat) scr_w2-crosshairSize/2.0f,
              (GLfloat) scr_h2-crosshairSize/2.0f,0.0f);
 // stvoruholnik
 glBegin(GL_QUADS); // glOrtho() vymeni Y
  glTexCoord2f(0.0f,0.0f); glVertex3f(0.0f,0.0f,0.0f);
  glTexCoord2f(0.0f,1.0f); glVertex3f(0.0f,crosshairSize,0.0f);
  glTexCoord2f(1.0f,1.0f); glVertex3f(crosshairSize,crosshairSize,0.0f);
  glTexCoord2f(1.0f,0.0f); glVertex3f(crosshairSize,0.0f,0.0f);
 glEnd();
 glPopMatrix(); // obnova modelview matice

 glColor4f(1.0f,1.0f,1.0f,1.0f); // biela

 glPopMatrix(); // obnova modelview matice
 glMatrixMode(GL_PROJECTION);
 glPopMatrix(); // obnova projection matice

 // aktualna je vzdy modelview matica
 glMatrixMode(GL_MODELVIEW);

 glDisable(GL_BLEND); // vypnutie priesvitnosti
 glPopAttrib();       // vyber ulozenych atributov - DEPTH_BUFFER
}

// vypocet TBN vektorov ****************************************************
void CInterface::TBN(int E,int Q,int W)
{
 vec3 v0,v1,v2;
 vec2 v0t,v1t,v2t;
 vec3 tan,bit,nor;

 v0.set(v_a[E+Q+0],v_a[E+Q+1],v_a[E+Q+2]); // 3 vrcholy
 v1.set(v_a[E+Q+3],v_a[E+Q+4],v_a[E+Q+5]);
 v2.set(v_a[E+Q+6],v_a[E+Q+7],v_a[E+Q+8]);
 if (E==0) { // kocka
  v0t.set(t_a[W+0],t_a[W+1]);              // 3 texturovacie suradnice
  v1t.set(t_a[W+2],t_a[W+3]);
  v2t.set(t_a[W+4],t_a[W+5]);
 }
 else {      // ostatne objekty
  v0t.set(t_a[72+W+0],t_a[72+W+1]);        // 3 texturovacie suradnice
  v1t.set(t_a[72+W+2],t_a[72+W+3]);
  v2t.set(t_a[72+W+4],t_a[72+W+5]);
 }
 mm.calc_tbn(v0,v1,v2,v0t,v1t,v2t,tan,bit,nor); // vypocet TBN
 tan_a[E+Q+0]=tan.x; tan_a[E+Q+1]=tan.y; tan_a[E+Q+2]=tan.z; // 1. vrchol
 tan_a[E+Q+3]=tan.x; tan_a[E+Q+4]=tan.y; tan_a[E+Q+5]=tan.z; // 2. vrchol
 tan_a[E+Q+6]=tan.x; tan_a[E+Q+7]=tan.y; tan_a[E+Q+8]=tan.z; // 3. vrchol
 bit_a[E+Q+0]=bit.x; bit_a[E+Q+1]=bit.y; bit_a[E+Q+2]=bit.z; // 1. vrchol
 bit_a[E+Q+3]=bit.x; bit_a[E+Q+4]=bit.y; bit_a[E+Q+5]=bit.z; // 2. vrchol
 bit_a[E+Q+6]=bit.x; bit_a[E+Q+7]=bit.y; bit_a[E+Q+8]=bit.z; // 3. vrchol
 nor_a[E+Q+0]=nor.x; nor_a[E+Q+1]=nor.y; nor_a[E+Q+2]=nor.z; // 1. vrchol
 nor_a[E+Q+3]=nor.x; nor_a[E+Q+4]=nor.y; nor_a[E+Q+5]=nor.z; // 2. vrchol
 nor_a[E+Q+6]=nor.x; nor_a[E+Q+7]=nor.y; nor_a[E+Q+8]=nor.z; // 3. vrchol
}

// najdenie prilahlych trojuholnikov a naplnenie pola indexov **************
void CInterface::search_indexes(int B,int E)
{
 int Tcount,Ti;
 vec3 v0,v1,v2;
 vec3 w0,w1,w2;
 vec3 dv;
 float d0,d1,d2,d3,d4,d5,d6,d7,d8;
 int e0,e1,e2;

 // B - index prveho prvku pola vrcholov objektu
 // E - index posledneho prvku pola vrcholov objektu
 Tcount=(E-B)/9; // pocet trojuholnikov objektu

 // index prveho vrcholu objektu
 Ti=B/3;

 printf("index prveho prvku: %i\n",B);
 printf("index posledneho prvku: %i\n",E);
 printf("index prveho vrcholu: %i\n",Ti);
 printf("pocet trojuholnikov: %i\n",Tcount);

 // pre vsetky trojuholniky
 loopi(Tcount) {
  // trojuholnik, ku ktoremu sa hladaju prilahle trojuholniky
  v0.set(v_a[B+i*9+0],v_a[B+i*9+1],v_a[B+i*9+2]);
  v1.set(v_a[B+i*9+3],v_a[B+i*9+4],v_a[B+i*9+5]);
  v2.set(v_a[B+i*9+6],v_a[B+i*9+7],v_a[B+i*9+8]);

  // indexy vrcholov trojuholnika
  i_a[i_ac+0]=Ti+i*3+0;
  i_a[i_ac+1]=0;
  i_a[i_ac+2]=Ti+i*3+1;
  i_a[i_ac+3]=0;
  i_a[i_ac+4]=Ti+i*3+2;
  i_a[i_ac+5]=0;

  // indikator najdenia trojuholnika k danej strane
  e0=0; e1=0; e2=0;

  // pre vsetky trojuholniky
  loopj(Tcount) {
   // prehladavaju sa vsetky ostatne trojuholniky  
   if (i==j) continue;
   // trojuholnik, ktory sa bude porovnavat
   w0.set(v_a[B+j*9+0],v_a[B+j*9+1],v_a[B+j*9+2]);
   w1.set(v_a[B+j*9+3],v_a[B+j*9+4],v_a[B+j*9+5]);
   w2.set(v_a[B+j*9+6],v_a[B+j*9+7],v_a[B+j*9+8]);
   // vzdialenost vrcholov trojuholnikov
   dv=w0-v0; d0=dv.magnitude();
   dv=w1-v0; d1=dv.magnitude();
   dv=w2-v0; d2=dv.magnitude();
   dv=w0-v1; d3=dv.magnitude();
   dv=w1-v1; d4=dv.magnitude();
   dv=w2-v1; d5=dv.magnitude();
   dv=w0-v2; d6=dv.magnitude();
   dv=w1-v2; d7=dv.magnitude();
   dv=w2-v2; d8=dv.magnitude();
   // ak trojuholnik prilieha k 1. strane trojuholnika
   if ((d2<0.001f && d4<0.001f) || (d1<0.001f && d5<0.001f)) { i_a[i_ac+1]=Ti+j*3+0; e0++; continue; }
   if ((d0<0.001f && d5<0.001f) || (d2<0.001f && d3<0.001f)) { i_a[i_ac+1]=Ti+j*3+1; e0++; continue; }
   if ((d1<0.001f && d3<0.001f) || (d4<0.001f && d0<0.001f)) { i_a[i_ac+1]=Ti+j*3+2; e0++; continue; }
   // ak trojuholnik prilieha k 2. strane trojuholnika 
   if ((d4<0.001f && d8<0.001f) || (d7<0.001f && d5<0.001f)) { i_a[i_ac+3]=Ti+j*3+0; e1++; continue; }
   if ((d3<0.001f && d8<0.001f) || (d5<0.001f && d6<0.001f)) { i_a[i_ac+3]=Ti+j*3+1; e1++; continue; }
   if ((d6<0.001f && d4<0.001f) || (d7<0.001f && d3<0.001f)) { i_a[i_ac+3]=Ti+j*3+2; e1++; continue; }
   // ak trojuholnik prilieha k 3. strane trojuholnika 
   if ((d1<0.001f && d8<0.001f) || (d2<0.001f && d7<0.001f)) { i_a[i_ac+5]=Ti+j*3+0; e2++; continue; }
   if ((d0<0.001f && d8<0.001f) || (d2<0.001f && d6<0.001f)) { i_a[i_ac+5]=Ti+j*3+1; e2++; continue; }
   if ((d0<0.001f && d7<0.001f) || (d1<0.001f && d6<0.001f)) { i_a[i_ac+5]=Ti+j*3+2; e2++; continue; }
  }

  /*if (e0==0 || e1==0 || e2==0) {
   printf("Nenasiel sa prilahly trojuholnik k jednej zo stran!\n");
  }*/

  i_ac+=6;
 }
}

// priprava objektov - kocka, gula, valec, ihlan, plochy *******************
void CInterface::objects_prepare(void)
{
 // pomocne premenne
 int q=0,w=0,e=0,f=0,r=0,s=0,t=0;
 float alfa[2],beta[2];
 float x[4],y[2],z[4];
 vec3 n[7];

 e=108; // pocet udajov patriacich kocke

 // kocka ------------------------------------------------------------------
 float tmp[]={-4.0f, 0.0f, 4.0f,-2.0f, 0.0f, 4.0f,-2.0f, 2.0f, 4.0f,
              -4.0f, 0.0f, 4.0f,-2.0f, 2.0f, 4.0f,-4.0f, 2.0f, 4.0f,
              -2.0f, 0.0f, 4.0f,-2.0f, 0.0f, 2.0f,-2.0f, 2.0f, 2.0f,
              -2.0f, 0.0f, 4.0f,-2.0f, 2.0f, 2.0f,-2.0f, 2.0f, 4.0f,
              -2.0f, 0.0f, 2.0f,-4.0f, 0.0f, 2.0f,-4.0f, 2.0f, 2.0f,
              -2.0f, 0.0f, 2.0f,-4.0f, 2.0f, 2.0f,-2.0f, 2.0f, 2.0f,
              -4.0f, 0.0f, 2.0f,-4.0f, 0.0f, 4.0f,-4.0f, 2.0f, 4.0f,
              -4.0f, 0.0f, 2.0f,-4.0f, 2.0f, 4.0f,-4.0f, 2.0f, 2.0f,
              -4.0f, 2.0f, 4.0f,-2.0f, 2.0f, 4.0f,-2.0f, 2.0f, 2.0f,
              -4.0f, 2.0f, 4.0f,-2.0f, 2.0f, 2.0f,-4.0f, 2.0f, 2.0f,
              -4.0f, 0.0f, 2.0f,-2.0f, 0.0f, 2.0f,-2.0f, 0.0f, 4.0f,
              -4.0f, 0.0f, 2.0f,-2.0f, 0.0f, 4.0f,-4.0f, 0.0f, 4.0f};
 loopi(e) v_a[i]=tmp[i];
 // farby
 for (int i=0;i<e;i+=3) { c_a[i]=0.5f; c_a[i+1]=0.0f; c_a[i+2]=0.0f; }
 // normaly
 float tmp1[]={ 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,
                0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,
                1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f,
                0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f,
               -1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,
               -1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,
                0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
                0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
                0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f,
                0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f, 0.0f,-1.0f, 0.0f};
 loopi(e) nor_a[i]=tmp1[i];
 // texturovacie suradnice
 float tmp2[]={0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 1.0f,
               0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f};
 loopi(72) t_a[i]=tmp2[i];
 // TBN pre kocku
 for (int i=0,j=0;i<e;i+=9,j+=6) TBN(0,i,j);

 // najdenie prilahlych trojuholnikov, naplnenie pola indexov
 i_ac=0; // pocitadlo - inkrementuje sa v ramci funkcie search_indexes()
 search_indexes(f,e+q);

 f=e+q; // zaciatok udajov gule

 // gula -------------------------------------------------------------------
 loopj(180) {
  if (j%segs) continue;
  loopi(360) {
   if (i%segs) continue;

   alfa[0]=DEG2RAD((float) i);        beta[0]=DEG2RAD((float) j);
   alfa[1]=DEG2RAD((float) (i+segs)); beta[1]=DEG2RAD((float) (j+segs));

   x[0]=3.0f+cos(alfa[0])*sin(beta[0]); x[1]=3.0f+cos(alfa[1])*sin(beta[0]);
   x[2]=3.0f+cos(alfa[1])*sin(beta[1]); x[3]=3.0f+cos(alfa[0])*sin(beta[1]);
   y[0]=1.0f+cos(beta[0]); y[1]=1.0f+cos(beta[1]);
   z[0]=3.0f+sin(alfa[0])*sin(beta[0]); z[1]=3.0f+sin(alfa[1])*sin(beta[0]);
   z[2]=3.0f+sin(alfa[1])*sin(beta[1]); z[3]=3.0f+sin(alfa[0])*sin(beta[1]);
   if (j!=0) { // nevytvara sa prvy z paru trojuholnikov na severnom pole
    v_a[e+q+0]=x[0]; v_a[e+q+1]=y[0]; v_a[e+q+2]=z[0]; // 1. vrchol
    v_a[e+q+3]=x[1]; v_a[e+q+4]=y[0]; v_a[e+q+5]=z[1]; // 2. vrchol
    v_a[e+q+6]=x[2]; v_a[e+q+7]=y[1]; v_a[e+q+8]=z[2]; // 3. vrchol
    c_a[e+q+0]=0.0f; c_a[e+q+1]=0.5f; c_a[e+q+2]=0.0f; // 1. vrchol
    c_a[e+q+3]=0.0f; c_a[e+q+4]=0.5f; c_a[e+q+5]=0.0f; // 2. vrchol
    c_a[e+q+6]=0.0f; c_a[e+q+7]=0.5f; c_a[e+q+8]=0.0f; // 3. vrchol
    t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=(float) j/90.0f;        // 1. vrchol
    t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=(float) j/90.0f;        // 2. vrchol
    t_a[72+w+4]=(float) (i+segs)/90.0f; t_a[72+w+5]=(float) (j+segs)/90.0f; // 3. vrchol
    TBN(e,q,w);
    q+=9; w+=6;
   }
   if (j!=(180-segs)) { // nevytvara sa druhy z paru trojuholnikov na juznom pole
    v_a[e+q+0]=x[0]; v_a[e+q+1]=y[0]; v_a[e+q+2]=z[0]; // 1. vrchol
    v_a[e+q+3]=x[2]; v_a[e+q+4]=y[1]; v_a[e+q+5]=z[2]; // 2. vrchol
    v_a[e+q+6]=x[3]; v_a[e+q+7]=y[1]; v_a[e+q+8]=z[3]; // 3. vrchol
    c_a[e+q+0]=0.0f; c_a[e+q+1]=0.5f; c_a[e+q+2]=0.0f; // 1. vrchol
    c_a[e+q+3]=0.0f; c_a[e+q+4]=0.5f; c_a[e+q+5]=0.0f; // 2. vrchol
    c_a[e+q+6]=0.0f; c_a[e+q+7]=0.5f; c_a[e+q+8]=0.0f; // 3. vrchol
    t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=(float) j/90.0f;        // 1. vrchol
    t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=(float) (j+segs)/90.0f; // 2. vrchol
    t_a[72+w+4]=(float) i/90.0f;        t_a[72+w+5]=(float) (j+segs)/90.0f; // 3. vrchol
    TBN(e,q,w);
    q+=9; w+=6;
   }
  }
 }
 // vyhladenie normal
 t=18*(360/segs);
 loopj(180) {
  if ((j==0) || (j%segs)) continue;
  loopi(360) {
   if (i%segs) continue;
   r=i+j*(360/segs);
   s=i+(j-segs)*(360/segs);
   if (i==0) {
    n[0].set(nor_a[f+r+0],   nor_a[f+r+1],   nor_a[f+r+2]);
    n[1].set(nor_a[f+r+9],   nor_a[f+r+10],  nor_a[f+r+11]);
    n[2].set(nor_a[f+r+t-15],nor_a[f+r+t-14],nor_a[f+r+t-13]);
    n[3].set(nor_a[f+s+t-6], nor_a[f+s+t-5], nor_a[f+s+t-4]);
    n[4].set(nor_a[f+s+t-12],nor_a[f+s+t-11],nor_a[f+s+t-10]);
    n[5].set(nor_a[f+s+15],  nor_a[f+s+16],  nor_a[f+s+17]);
    n[6]=(n[0]+n[1]+n[2]+n[3]+n[4]+n[5])/6.0f; n[6].normalize();
    nor_a[f+r+0]=   n[6].x; nor_a[f+r+1]=   n[6].y; nor_a[f+r+2]=   n[6].z;
    nor_a[f+r+9]=   n[6].x; nor_a[f+r+10]=  n[6].y; nor_a[f+r+11]=  n[6].z;
    nor_a[f+r+t-15]=n[6].x; nor_a[f+r+t-14]=n[6].y; nor_a[f+r+t-13]=n[6].z;
    nor_a[f+s+t-6]= n[6].x; nor_a[f+s+t-5]= n[6].y; nor_a[f+s+t-4]= n[6].z;
    nor_a[f+s+t-12]=n[6].x; nor_a[f+s+t-11]=n[6].y; nor_a[f+s+t-10]=n[6].z;
    nor_a[f+s+15]=  n[6].x; nor_a[f+s+16]=  n[6].y; nor_a[f+s+17]=  n[6].z;
   }
   else {
    n[0].set(nor_a[f+r+0], nor_a[f+r+1], nor_a[f+r+2]);
    n[1].set(nor_a[f+r+9], nor_a[f+r+10],nor_a[f+r+11]);
    n[2].set(nor_a[f+r-15],nor_a[f+r-14],nor_a[f+r-13]);
    n[3].set(nor_a[f+s-6], nor_a[f+s-5], nor_a[f+s-4]);
    n[4].set(nor_a[f+s-12],nor_a[f+s-11],nor_a[f+s-10]);
    n[5].set(nor_a[f+s+15],nor_a[f+s+16],nor_a[f+s+17]);
    n[6]=(n[0]+n[1]+n[2]+n[3]+n[4]+n[5])/6.0f; n[6].normalize();
    nor_a[f+r+0]= n[6].x; nor_a[f+r+1]= n[6].y; nor_a[f+r+2]= n[6].z;
    nor_a[f+r+9]= n[6].x; nor_a[f+r+10]=n[6].y; nor_a[f+r+11]=n[6].z;
    nor_a[f+r-15]=n[6].x; nor_a[f+r-14]=n[6].y; nor_a[f+r-13]=n[6].z;
    nor_a[f+s-6]= n[6].x; nor_a[f+s-5]= n[6].y; nor_a[f+s-4]= n[6].z;
    nor_a[f+s-12]=n[6].x; nor_a[f+s-11]=n[6].y; nor_a[f+s-10]=n[6].z;
    nor_a[f+s+15]=n[6].x; nor_a[f+s+16]=n[6].y; nor_a[f+s+17]=n[6].z;
   }
  }
 }

 // najdenie prilahlych trojuholnikov, naplnenie pola indexov
 search_indexes(f,e+q);

 f=e+q; // zaciatok udajov valca

 // valec ------------------------------------------------------------------
 loopi(360) {
  if (i%segs) continue;

  alfa[0]=DEG2RAD((float) i);
  alfa[1]=DEG2RAD((float) (i+segs));

  x[0]=3.0f+cos(alfa[0]); x[1]=3.0f+cos(alfa[1]);
  y[0]=0.0f; y[1]=2.0f;
  z[0]=-3.0f-sin(alfa[0]); z[1]=-3.0f-sin(alfa[1]);

  v_a[e+q+0]=x[0]; v_a[e+q+1]=y[0]; v_a[e+q+2]=z[0]; // 1. vrchol
  v_a[e+q+3]=x[1]; v_a[e+q+4]=y[0]; v_a[e+q+5]=z[1]; // 2. vrchol
  v_a[e+q+6]=x[1]; v_a[e+q+7]=y[1]; v_a[e+q+8]=z[1]; // 3. vrchol
  c_a[e+q+0]=0.0f; c_a[e+q+1]=0.0f; c_a[e+q+2]=0.5f; // 1. vrchol
  c_a[e+q+3]=0.0f; c_a[e+q+4]=0.0f; c_a[e+q+5]=0.5f; // 2. vrchol
  c_a[e+q+6]=0.0f; c_a[e+q+7]=0.0f; c_a[e+q+8]=0.5f; // 3. vrchol
  t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=0.0f; // 2. vrchol
  t_a[72+w+4]=(float) (i+segs)/90.0f; t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;

  v_a[e+q+0]=x[0]; v_a[e+q+1]=y[0]; v_a[e+q+2]=z[0]; // 1. vrchol
  v_a[e+q+3]=x[1]; v_a[e+q+4]=y[1]; v_a[e+q+5]=z[1]; // 2. vrchol
  v_a[e+q+6]=x[0]; v_a[e+q+7]=y[1]; v_a[e+q+8]=z[0]; // 3. vrchol
  c_a[e+q+0]=0.0f; c_a[e+q+1]=0.0f; c_a[e+q+2]=0.5f; // 1. vrchol
  c_a[e+q+3]=0.0f; c_a[e+q+4]=0.0f; c_a[e+q+5]=0.5f; // 2. vrchol
  c_a[e+q+6]=0.0f; c_a[e+q+7]=0.0f; c_a[e+q+8]=0.5f; // 3. vrchol
  t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=1.0f; // 2. vrchol
  t_a[72+w+4]=(float) i/90.0f;        t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;

  v_a[e+q+0]=x[0]; v_a[e+q+1]=2.0f; v_a[e+q+2]=z[0];  // 1. vrchol
  v_a[e+q+3]=x[1]; v_a[e+q+4]=2.0f; v_a[e+q+5]=z[1];  // 2. vrchol
  v_a[e+q+6]=3.0f; v_a[e+q+7]=2.0f; v_a[e+q+8]=-3.0f; // 3. vrchol
  c_a[e+q+0]=0.0f; c_a[e+q+1]=0.0f; c_a[e+q+2]=0.5f; // 1. vrchol
  c_a[e+q+3]=0.0f; c_a[e+q+4]=0.0f; c_a[e+q+5]=0.5f; // 2. vrchol
  c_a[e+q+6]=0.0f; c_a[e+q+7]=0.0f; c_a[e+q+8]=0.5f; // 3. vrchol
  t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=0.0f; // 2. vrchol
  t_a[72+w+4]=(float) (i/90);         t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;

  v_a[e+q+0]=x[1]; v_a[e+q+1]=0.0f; v_a[e+q+2]=z[1];  // 1. vrchol
  v_a[e+q+3]=x[0]; v_a[e+q+4]=0.0f; v_a[e+q+5]=z[0];  // 2. vrchol
  v_a[e+q+6]=3.0f; v_a[e+q+7]=0.0f; v_a[e+q+8]=-3.0f; // 3. vrchol
  c_a[e+q+0]=0.0f; c_a[e+q+1]=0.0f; c_a[e+q+2]=0.5f; // 1. vrchol
  c_a[e+q+3]=0.0f; c_a[e+q+4]=0.0f; c_a[e+q+5]=0.5f; // 2. vrchol
  c_a[e+q+6]=0.0f; c_a[e+q+7]=0.0f; c_a[e+q+8]=0.5f; // 3. vrchol
  t_a[72+w+0]=(float) (i+segs)/90.0f; t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) i/90.0f;        t_a[72+w+3]=0.0f; // 2. vrchol
  t_a[72+w+4]=(float) (i/90);         t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;
 }
 // vyhladenie normal
 t=36*(360/segs);
 for (int i=0;i<36*(360/segs);i+=36) {
  if (i==0) {
   n[0].set(nor_a[f+i+0],   nor_a[f+i+1],   nor_a[f+i+2]);
   n[1].set(nor_a[f+i+9],   nor_a[f+i+10],  nor_a[f+i+11]);
   n[2].set(nor_a[f+i+t-33],nor_a[f+i+t-32],nor_a[f+i+t-31]);
   n[3].set(nor_a[f+i+15],  nor_a[f+i+16],  nor_a[f+i+17]);
   n[4].set(nor_a[f+i+t-30],nor_a[f+i+t-29],nor_a[f+i+t-28]);
   n[5].set(nor_a[f+i+t-24],nor_a[f+i+t-23],nor_a[f+i+t-22]);
   n[6]=(n[0]+n[1]+n[2])/3.0f; n[6].normalize();
   nor_a[f+i+0]=   n[6].x; nor_a[f+i+1]=   n[6].y; nor_a[f+i+2]=   n[6].z;
   nor_a[f+i+9]=   n[6].x; nor_a[f+i+10]=  n[6].y; nor_a[f+i+11]=  n[6].z;
   nor_a[f+i+t-33]=n[6].x; nor_a[f+i+t-32]=n[6].y; nor_a[f+i+t-31]=n[6].z;
   n[6]=(n[3]+n[4]+n[5])/3.0f; n[6].normalize();
   nor_a[f+i+15]=  n[6].x; nor_a[f+i+16]=  n[6].y; nor_a[f+i+17]=  n[6].z;
   nor_a[f+i+t-30]=n[6].x; nor_a[f+i+t-29]=n[6].y; nor_a[f+i+t-28]=n[6].z;
   nor_a[f+i+t-24]=n[6].x; nor_a[f+i+t-23]=n[6].y; nor_a[f+i+t-22]=n[6].z;
  }
  else {
   n[0].set(nor_a[f+i+0], nor_a[f+i+1], nor_a[f+i+2]);
   n[1].set(nor_a[f+i+9], nor_a[f+i+10],nor_a[f+i+11]);
   n[2].set(nor_a[f+i-33],nor_a[f+i-32],nor_a[f+i-31]);
   n[3].set(nor_a[f+i+15],nor_a[f+i+16],nor_a[f+i+17]);
   n[4].set(nor_a[f+i-30],nor_a[f+i-29],nor_a[f+i-28]);
   n[5].set(nor_a[f+i-24],nor_a[f+i-23],nor_a[f+i-22]);
   n[6]=(n[0]+n[1]+n[2])/3.0f; n[6].normalize();
   nor_a[f+i+0]= n[6].x; nor_a[f+i+1]= n[6].y; nor_a[f+i+2]= n[6].z;
   nor_a[f+i+9]= n[6].x; nor_a[f+i+10]=n[6].y; nor_a[f+i+11]=n[6].z;
   nor_a[f+i-33]=n[6].x; nor_a[f+i-32]=n[6].y; nor_a[f+i-31]=n[6].z;
   n[6]=(n[3]+n[4]+n[5])/3.0f; n[6].normalize();
   nor_a[f+i+15]=n[6].x; nor_a[f+i+16]=n[6].y; nor_a[f+i+17]=n[6].z;
   nor_a[f+i-30]=n[6].x; nor_a[f+i-29]=n[6].y; nor_a[f+i-28]=n[6].z;
   nor_a[f+i-24]=n[6].x; nor_a[f+i-23]=n[6].y; nor_a[f+i-22]=n[6].z;
  }
 }

 // najdenie prilahlych trojuholnikov, naplnenie pola indexov
 search_indexes(f,e+q);

 f=e+q; // zaciatok udajov ihlanu

 // ihlan ------------------------------------------------------------------
 loopi(360) {
  if (i%segs) continue;

  alfa[0]=DEG2RAD((float) i);
  alfa[1]=DEG2RAD((float) (i+segs));

  x[0]=-3.0f+cos(alfa[0]); x[1]=-3.0f+cos(alfa[1]);
  y[0]=0.0f; y[1]=2.0f;
  z[0]=-3.0f-sin(alfa[0]); z[1]=-3.0f-sin(alfa[1]);

  v_a[e+q+0]=x[0];  v_a[e+q+1]=y[0]; v_a[e+q+2]=z[0];  // 1. vrchol
  v_a[e+q+3]=x[1];  v_a[e+q+4]=y[0]; v_a[e+q+5]=z[1];  // 2. vrchol
  v_a[e+q+6]=-3.0f; v_a[e+q+7]=y[1]; v_a[e+q+8]=-3.0f; // 3. vrchol
  c_a[e+q+0]=0.5f; c_a[e+q+1]=0.5f; c_a[e+q+2]=0.0f; // 1. vrchol
  c_a[e+q+3]=0.5f; c_a[e+q+4]=0.5f; c_a[e+q+5]=0.0f; // 2. vrchol
  c_a[e+q+6]=0.5f; c_a[e+q+7]=0.5f; c_a[e+q+8]=0.0f; // 3. vrchol
  t_a[72+w+0]=(float) i/90.0f;        t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) (i+segs)/90.0f; t_a[72+w+3]=0.0f; // 2. vrchol
  t_a[72+w+4]=(float) (i/90);         t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;

  v_a[e+q+0]=x[1];  v_a[e+q+1]=0.0f; v_a[e+q+2]=z[1];  // 1. vrchol
  v_a[e+q+3]=x[0];  v_a[e+q+4]=0.0f; v_a[e+q+5]=z[0];  // 2. vrchol
  v_a[e+q+6]=-3.0f; v_a[e+q+7]=0.0f; v_a[e+q+8]=-3.0f; // 3. vrchol
  c_a[e+q+0]=0.5f; c_a[e+q+1]=0.5f; c_a[e+q+2]=0.0f; // 1. vrchol
  c_a[e+q+3]=0.5f; c_a[e+q+4]=0.5f; c_a[e+q+5]=0.0f; // 2. vrchol
  c_a[e+q+6]=0.5f; c_a[e+q+7]=0.5f; c_a[e+q+8]=0.0f; // 3. vrchol
  t_a[72+w+0]=(float) (i+segs)/90.0f; t_a[72+w+1]=0.0f; // 1. vrchol
  t_a[72+w+2]=(float) i/90.0f;        t_a[72+w+3]=0.0f; // 2. vrchol
  t_a[72+w+4]=(float) (i/90);         t_a[72+w+5]=1.0f; // 3. vrchol
  TBN(e,q,w);
  q+=9; w+=6;
 }
 // vyhladenie normal
 t=18*(360/segs);
 for (int i=0;i<18*(360/segs);i+=18) {
  if (i==0) {
   n[0].set(nor_a[f+i+0],   nor_a[f+i+1],   nor_a[f+i+2]);
   n[1].set(nor_a[f+i+t-15],nor_a[f+i+t-14],nor_a[f+i+t-13]);
   n[6]=(n[0]+n[1])/2.0f; n[6].normalize();
   nor_a[f+i+0]=   n[6].x; nor_a[f+i+1]=   n[6].y; nor_a[f+i+2]=   n[6].z;
   nor_a[f+i+t-15]=n[6].x; nor_a[f+i+t-14]=n[6].y; nor_a[f+i+t-13]=n[6].z;
   nor_a[f+i+6]=   0.0f; nor_a[f+i+7]=   1.0f; nor_a[f+i+8]=   0.0f;
   nor_a[f+i+t-12]=0.0f; nor_a[f+i+t-11]=1.0f; nor_a[f+i+t-10]=0.0f;
  }
  else {
   n[0].set(nor_a[f+i+0], nor_a[f+i+1], nor_a[f+i+2]);
   n[1].set(nor_a[f+i-15],nor_a[f+i-14],nor_a[f+i-13]);
   n[2].set(nor_a[f+i+6], nor_a[f+i+7], nor_a[f+i+8]);
   n[3].set(nor_a[f+i-12],nor_a[f+i-11],nor_a[f+i-10]);
   n[6]=(n[0]+n[1])/2.0f; n[6].normalize();
   nor_a[f+i+0]= n[6].x; nor_a[f+i+1]= n[6].y; nor_a[f+i+2]= n[6].z;
   nor_a[f+i-15]=n[6].x; nor_a[f+i-14]=n[6].y; nor_a[f+i-13]=n[6].z;
   nor_a[f+i+6]= 0.0f; nor_a[f+i+7]= 1.0f; nor_a[f+i+8]= 0.0f;
   nor_a[f+i-12]=0.0f; nor_a[f+i-11]=1.0f; nor_a[f+i-10]=0.0f;
  }
 }

 // najdenie prilahlych trojuholnikov, naplnenie pola indexov
 search_indexes(f,e+q);

 f=e+q;

 // plochy -----------------------------------------------------------------
 v_a[e+q+0]=-5.0f; v_a[e+q+1]=0.0f; v_a[e+q+2]= 5.0f; // 1. vrchol
 v_a[e+q+3]= 5.0f; v_a[e+q+4]=0.0f; v_a[e+q+5]= 5.0f; // 2. vrchol
 v_a[e+q+6]= 5.0f; v_a[e+q+7]=0.0f; v_a[e+q+8]=-5.0f; // 3. vrchol
 c_a[e+q+0]=0.2f; c_a[e+q+1]=0.2f; c_a[e+q+2]=0.2f; // 1. vrchol
 c_a[e+q+3]=0.2f; c_a[e+q+4]=0.2f; c_a[e+q+5]=0.2f; // 2. vrchol
 c_a[e+q+6]=0.2f; c_a[e+q+7]=0.2f; c_a[e+q+8]=0.2f; // 3. vrchol
 t_a[72+w+0]=0.0f; t_a[72+w+1]=0.0f; // 1. vrchol
 t_a[72+w+2]=1.0f; t_a[72+w+3]=0.0f; // 2. vrchol
 t_a[72+w+4]=1.0f; t_a[72+w+5]=1.0f; // 3. vrchol
 TBN(e,q,w);
 q+=9; w+=6;

 v_a[e+q+0]=-5.0f; v_a[e+q+1]=0.0f; v_a[e+q+2]= 5.0f; // 1. vrchol
 v_a[e+q+3]= 5.0f; v_a[e+q+4]=0.0f; v_a[e+q+5]=-5.0f; // 2. vrchol
 v_a[e+q+6]=-5.0f; v_a[e+q+7]=0.0f; v_a[e+q+8]=-5.0f; // 3. vrchol
 c_a[e+q+0]=0.2f; c_a[e+q+1]=0.2f; c_a[e+q+2]=0.2f; // 1. vrchol
 c_a[e+q+3]=0.2f; c_a[e+q+4]=0.2f; c_a[e+q+5]=0.2f; // 2. vrchol
 c_a[e+q+6]=0.2f; c_a[e+q+7]=0.2f; c_a[e+q+8]=0.2f; // 3. vrchol
 t_a[72+w+0]=0.0f; t_a[72+w+1]=0.0f; // 1. vrchol
 t_a[72+w+2]=1.0f; t_a[72+w+3]=1.0f; // 2. vrchol
 t_a[72+w+4]=0.0f; t_a[72+w+5]=1.0f; // 3. vrchol
 TBN(e,q,w);
 q+=9; w+=6;

 v_a[e+q+0]=-5.0f; v_a[e+q+1]=0.0f;  v_a[e+q+2]=-5.0f; // 1. vrchol
 v_a[e+q+3]= 5.0f; v_a[e+q+4]=0.0f;  v_a[e+q+5]=-5.0f; // 2. vrchol
 v_a[e+q+6]= 5.0f; v_a[e+q+7]=10.0f; v_a[e+q+8]=-5.0f; // 3. vrchol
 c_a[e+q+0]=0.2f; c_a[e+q+1]=0.2f; c_a[e+q+2]=0.2f; // 1. vrchol
 c_a[e+q+3]=0.2f; c_a[e+q+4]=0.2f; c_a[e+q+5]=0.2f; // 2. vrchol
 c_a[e+q+6]=0.2f; c_a[e+q+7]=0.2f; c_a[e+q+8]=0.2f; // 3. vrchol
 t_a[72+w+0]=0.0f; t_a[72+w+1]=0.0f; // 1. vrchol
 t_a[72+w+2]=1.0f; t_a[72+w+3]=0.0f; // 2. vrchol
 t_a[72+w+4]=1.0f; t_a[72+w+5]=1.0f; // 3. vrchol
 TBN(e,q,w);
 q+=9; w+=6;

 v_a[e+q+0]=-5.0f; v_a[e+q+1]=0.0f; v_a[e+q+2]=-5.0f; // 1. vrchol
 v_a[e+q+3]= 5.0f; v_a[e+q+4]=10.0f; v_a[e+q+5]=-5.0f; // 2. vrchol
 v_a[e+q+6]=-5.0f; v_a[e+q+7]=10.0f; v_a[e+q+8]=-5.0f; // 3. vrchol
 c_a[e+q+0]=0.2f; c_a[e+q+1]=0.2f; c_a[e+q+2]=0.2f; // 1. vrchol
 c_a[e+q+3]=0.2f; c_a[e+q+4]=0.2f; c_a[e+q+5]=0.2f; // 2. vrchol
 c_a[e+q+6]=0.2f; c_a[e+q+7]=0.2f; c_a[e+q+8]=0.2f; // 3. vrchol
 t_a[72+w+0]=0.0f; t_a[72+w+1]=0.0f; // 1. vrchol
 t_a[72+w+2]=1.0f; t_a[72+w+3]=1.0f; // 2. vrchol
 t_a[72+w+4]=0.0f; t_a[72+w+5]=1.0f; // 3. vrchol
 TBN(e,q,w);
 q+=9; w+=6;

 // najdenie prilahlych trojuholnikov, naplnenie pola indexov
 //search_indexes(f,e+q);

 objs.vrts_num=e/3+q/3; // pocet vrcholov kocka+gula+valec+ihlan+plochy

 // indexy - pre standardne triangles sa pole naplni tu
 // pre triangles_adjacency sa pole naplna priebezne po vytvoreni objektu
 // v ramci funkcie search_indexes()
 //loopi(objs.vrts_num) i_a[i]=i;

 printf("pocet vrcholov: %i\n",objs.vrts_num);
 printf("pocet vrcholov pre triangles_adjacency: %i\n",i_ac);
 printf("pocet trojuholnikov: %i\n",objs.vrts_num/3);

 // Vertex Array Object
 glGenVertexArrays(1,&objects_vao);
 glBindVertexArray(objects_vao);

 // priprava objektov - kocka, gula, valec, ihlan, plochy
 objs.buffers_prepare();
 // nahratie udajov do zasobnikov
 objs.vertices->load(sizeof(v_a),v_a);
 objs.colors->load(sizeof(c_a),c_a);
 objs.tangents->load(sizeof(tan_a),tan_a);
 objs.bitangents->load(sizeof(bit_a),bit_a);
 objs.normals->load(sizeof(nor_a),nor_a);
 objs.texcoords->load(sizeof(t_a),t_a);
 objs.indices->load(sizeof(i_a),i_a);

#ifdef OGL20
 // vrcholy
 glEnableVertexAttribArray(0); // povolenie poli atributov
 // farby
 glEnableVertexAttribArray(1);
 // tangenty
 glEnableVertexAttribArray(2);
 // bitangenty
 glEnableVertexAttribArray(3);
 // normaly
 glEnableVertexAttribArray(4);
 // texturovacie suradnice
 glEnableVertexAttribArray(8);
#else
 // vrcholy
 glEnableVertexAttribArrayARB(0);
 // farby
 glEnableVertexAttribArrayARB(1);
 // tangenty
 glEnableVertexAttribArrayARB(2);
 // bitangenty
 glEnableVertexAttribArrayARB(3);
 // normaly
 glEnableVertexAttribArrayARB(4);
 // texturovacie suradnice
 glEnableVertexAttribArrayARB(8);
#endif

 // nastavenie ukazovatelov na udaje objektov
 objects_bind();

 glBindVertexArray(0);

 // INFO: bez shaderov
/*
 glEnableClientState(GL_VERTEX_ARRAY);
 glVertexPointer(3,GL_FLOAT,0,v_a);
 glEnableClientState(GL_COLOR_ARRAY);
 glColorPointer(3,GL_FLOAT,0,c_a);
*/
}

// nastavenie ukazovatelov na udaje objektov *******************************
void CInterface::objects_bind(void)
{
#ifdef OGL20
 // vrcholy
 objs.vertices->bind();                            // pripojenie buffra
 glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,0,0); // ukazovatel na udaje
 // farby
 objs.colors->bind();
 glVertexAttribPointer(1,3,GL_FLOAT,GL_FALSE,0,0);
 // tangenty
 objs.tangents->bind();
 glVertexAttribPointer(2,3,GL_FLOAT,GL_FALSE,0,0);
 // bitangenty
 objs.bitangents->bind();
 glVertexAttribPointer(3,3,GL_FLOAT,GL_FALSE,0,0);
 // normaly
 objs.normals->bind();
 glVertexAttribPointer(4,3,GL_FLOAT,GL_FALSE,0,0);
 // texturovacie suradnice
 objs.texcoords->bind();
 glVertexAttribPointer(8,2,GL_FLOAT,GL_FALSE,0,0);
#else
 // vrcholy
 objs.vertices->bind();
 glVertexAttribPointerARB(0,3,GL_FLOAT,GL_FALSE,0,0);
 // farby
 objs.colors->bind();
 glVertexAttribPointerARB(1,3,GL_FLOAT,GL_FALSE,0,0);
 // tangenty
 objs.tangents->bind();
 glVertexAttribPointerARB(2,3,GL_FLOAT,GL_FALSE,0,0);
 // bitangenty
 objs.bitangents->bind();
 glVertexAttribPointerARB(3,3,GL_FLOAT,GL_FALSE,0,0);
 // normaly
 objs.normals->bind();
 glVertexAttribPointerARB(4,3,GL_FLOAT,GL_FALSE,0,0);
 // texturovacie suradnice
 objs.texcoords->bind();
 glVertexAttribPointerARB(8,2,GL_FLOAT,GL_FALSE,0,0);
#endif
}

// odpojenie buffrov *******************************************************
void CInterface::objects_unbind(void)
{
 objs.vertices->unbind();
 objs.colors->unbind();
 objs.tangents->unbind();
 objs.bitangents->unbind();
 objs.normals->unbind();
 objs.texcoords->unbind();
}

// vykreslanie objektov - kocka, gula, valec, ihlan, plochy ****************
void CInterface::objects_draw(void)
{
 glBindVertexArray(objects_vao);

 // vsetky objekty
 if (but_text_id!=16) glDrawArrays(GL_TRIANGLES,0,objs.vrts_num);
 else glDrawElements(GL_TRIANGLES_ADJACENCY,i_ac,GL_UNSIGNED_INT,0);
 // iba plochy
 //glDrawArrays(GL_TRIANGLES,(4824-36)/3,12);

 glBindVertexArray(0);
}

// mys je/nie je na tlacidle ***********************************************
bool CInterface::on_button(float X,float Y)
{
 // pre vsetky tlacidla
 loopi(BUTNUM) {
  // ak je mys na tlacidle
  if (but[i].on_button(X,Y)) { but_id=i; return(true); }
 }
 // mys nie je na tlacidle
 return(false);
}

// zmena udajov tlacidiel **************************************************
void CInterface::button_update(float X)
{
 // zmena udajov konkretneho tlacidla
 old_but_data=but[but_id].get_data();
 but[but_id].update(X/200.0f);
 new_but_data=but[but_id].get_data();
 // zmena parametrov osvetlenia podla zmenenych udajov tlacidiel
 if (but_text_id!=10 && but_text_id!=15) { // ak nie je zvoleny HDR shader, alebo triangle_shrink
  switch (but_id) {
   case 0: lsrcparam.ambient.x=new_but_data; break;
   case 1: lsrcparam.ambient.y=new_but_data; break;
   case 2: lsrcparam.ambient.z=new_but_data; break;
   case 3: lsrcparam.ambient.w=new_but_data; break;
   case 4: lsrcparam.diffuse.x=new_but_data; break;
   case 5: lsrcparam.diffuse.y=new_but_data; break;
   case 6: lsrcparam.diffuse.z=new_but_data; break;
   case 7: lsrcparam.diffuse.w=new_but_data; break;
   case 8: lsrcparam.specular.x=new_but_data; break;
   case 9: lsrcparam.specular.y=new_but_data; break;
   case 10: lsrcparam.specular.z=new_but_data; break;
   case 11: lsrcparam.specular.w=new_but_data; break;
   case 12: lsrcparam.position.x=new_but_data; break;
   case 13: lsrcparam.position.y=new_but_data; break;
   case 14: lsrcparam.position.z=new_but_data; break;
   case 15: lsrcparam.position.w=new_but_data; break;
   case 16: lsrcparam.spotDirection.x=new_but_data; break;
   case 17: lsrcparam.spotDirection.y=new_but_data; break;
   case 18: lsrcparam.spotDirection.z=new_but_data; break;
   case 19: lsrcparam.spotExponent=new_but_data; break;
   case 20: lsrcparam.spotCutoff=new_but_data; break;
   case 21: lsrcparam.constantAttenuation=new_but_data; break;
   case 22: lsrcparam.linearAttenuation=new_but_data; break;
   case 23: lsrcparam.quadraticAttenuation=new_but_data; break;
   case 24: lmodelparam.ambient.x=new_but_data; break;
   case 25: lmodelparam.ambient.y=new_but_data; break;
   case 26: lmodelparam.ambient.z=new_but_data; break;
   case 27: lmodelparam.ambient.w=new_but_data; break;
   case 28: mparam.ambient.x=new_but_data; break;
   case 29: mparam.ambient.y=new_but_data; break;
   case 30: mparam.ambient.z=new_but_data; break;
   case 31: mparam.ambient.w=new_but_data; break;
   case 32: mparam.diffuse.x=new_but_data; break;
   case 33: mparam.diffuse.y=new_but_data; break;
   case 34: mparam.diffuse.z=new_but_data; break;
   case 35: mparam.diffuse.w=new_but_data; break;
   case 36: mparam.specular.x=new_but_data; break;
   case 37: mparam.specular.y=new_but_data; break;
   case 38: mparam.specular.z=new_but_data; break;
   case 39: mparam.specular.w=new_but_data; break;
   case 40: mparam.emission.x=new_but_data; break;
   case 41: mparam.emission.y=new_but_data; break;
   case 42: mparam.emission.z=new_but_data; break;
   case 43: mparam.emission.w=new_but_data; break;
   case 44: mparam.shininess=new_but_data; break;
   case 45: // zmena indexu textoveho udaju tlacidla = zmena pouziteho shadera
            but_text_id=(int) new_but_data;
            set_buttons(); // nastavenie parametrov a udajov tlacidiel menu
            // zmena textoveho udaju tlacidla
            but[45].set_data(new_but_data,but_text[but_text_id]);
            if (but_text_id==5) mparam.diffuse.set(0.6f,0.3f,0.1f,1.0f);
            else mparam.diffuse.set(0.8f,0.4f,0.4f,1.0f);
            but[32].set_data(mparam.diffuse.x); but[33].set_data(mparam.diffuse.y);
            but[34].set_data(mparam.diffuse.z); but[35].set_data(mparam.diffuse.w);
            break;
   case 46: // reload shadera
            but_id=-1; // reload shadera sa vykona iba 1x
            switch (but_text_id) {
             case 0: break;
             case 1: break;
             case 2: break;
             case 3: break;
             case 4: break;
             case 5: wood_shader.reload(0);

                     wood_shader.on();
                     wood_shader.bindattriblocation(0,"in_pos"); // vrcholy
                     wood_shader.bindattriblocation(4,"in_nor"); // normaly
                     wood_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                     wood_shader.uniform3F("lightWoodColor",0.6f,0.3f,0.1f);
                     wood_shader.uniform3F("darkWoodColor",0.4f,0.2f,0.07f);
                     wood_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                     wood_shader.off();
                     break;
             case 6: nm_shader.reload(0);

                     nm_shader.on();
                     nm_shader.bindattriblocation(0,"in_pos"); // vrcholy
                     nm_shader.bindattriblocation(2,"in_tan"); // tangenty
                     nm_shader.bindattriblocation(3,"in_bit"); // bitangenty
                     nm_shader.bindattriblocation(4,"in_nor"); // normaly
                     nm_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                     nm_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                     nm_shader.uniformI("nm_tex_unit",1);      // normalmap texturovacia jednotka
                     nm_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                     nm_shader.off();
                     break;
             case 7: px_shader.reload(0);

                     px_shader.on();
                     px_shader.bindattriblocation(0,"in_pos"); // vrcholy
                     px_shader.bindattriblocation(2,"in_tan"); // tangenty
                     px_shader.bindattriblocation(3,"in_bit"); // bitangenty
                     px_shader.bindattriblocation(4,"in_nor"); // normaly
                     px_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                     px_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                     px_shader.uniformI("hm_tex_unit",1);      // heightmap texturovacia jednotka
                     px_shader.uniformI("nm_tex_unit",2);      // normalmap texturovacia jednotka
                     px_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                     px_shader.off();
                     break;
             case 8: water_shader.reload(0);

                     water_shader.on();
                     water_shader.bindattriblocation(0,"in_pos"); // vrcholy
                     water_shader.bindattriblocation(2,"in_tan"); // tangenty
                     water_shader.bindattriblocation(3,"in_bit"); // bitangenty
                     water_shader.bindattriblocation(4,"in_nor"); // normaly
                     water_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                     water_shader.uniformI("rfl_tex_unit",0);     // reflection texturovacia jednotka
                     water_shader.uniformI("rfr_tex_unit",1);     // refraction texturovacia jednotka
                     water_shader.uniformI("depth_tex_unit",2);   // depth texturovacia jednotka
                     water_shader.uniformI("nm_tex_unit",3);      // normalmap texturovacia jednotka
                     water_shader.uniformI("dudv_tex_unit",4);    // dudv texturovacia jednotka
                     water_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                     water_shader.off();
                     break;
             case 11: ssao_shader[1].reload(0);

                      ssao_shader[1].on();
                      ssao_shader[1].bindattriblocation(0,"in_pos"); // vrcholy
                      ssao_shader[1].uniformI("depth_buf_unit",0);   // texturovacia jednotka pre hlbky
                      ssao_shader[1].uniformI("normal_buf_unit",1);  // texturovacia jednotka pre normaly
                      ssao_shader[1].uniformI("rndvec_buf_unit",2);  // texturovacia jednotka pre nahodne orientovane vektory
                      ssao_shader[1].uniformI("base_tex_unit",3);    // texturovacia jednotka pre farby
                      ssao_shader[1].uniformFV("samples",16,ssao_samples); // vzorky
                      ssao_shader[1].uniformmatrixFV("sbproj",1,sbproj.get_data());
                      ssao_shader[1].uniformmatrixFV("isbproj",1,isbproj.get_data());
                      ssao_shader[1].off();
                      break;
             case 12: shadow_shader.reload(0);

                      shadow_shader.on();
                      shadow_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      shadow_shader.bindattriblocation(4,"in_nor"); // normaly
                      shadow_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      shadow_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      shadow_shader.uniformI("sm_tex_unit",1);      // shadowmap texturovacia jednotka
                      shadow_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      shadow_shader.off();
                      break;
             case 13: null_shader.reload(0);

                      null_shader.on();
                      null_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      null_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      null_shader.off();

                      raw_shader.reload(0);

                      raw_shader.on();
                      raw_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      raw_shader.bindattriblocation(2,"in_tan"); // tangenty
                      raw_shader.bindattriblocation(3,"in_bit"); // bitangenty
                      raw_shader.bindattriblocation(4,"in_nor"); // normaly
                      raw_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      raw_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      raw_shader.uniformI("nm_tex_unit",1);      // normalmap texturovacia jednotka
                      raw_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      raw_shader.off();

                      def_shader.reload(0);

                      def_shader.on();
                      def_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      def_shader.uniformI("depth_tex_unit",0);   // texturovacia jednotka pre hlbky
                      def_shader.uniformI("albedo_tex_unit",1);  // texturovacia jednotka pre albedo
                      def_shader.uniformI("normal_tex_unit",2);  // texturovacia jednotka pre normaly
                      def_shader.uniformmatrixFV("sbproj",1,sbproj.get_data());
                      def_shader.uniformmatrixFV("isbproj",1,isbproj.get_data());
                      // naplnenie pola difuznymi komponentami 16-tich svetelnych zdrojov
                      a=0;
                      loopi(16) {
                       lightpos[a+0]=mlight[i].diffuse.x; lightpos[a+1]=mlight[i].diffuse.y;
                       lightpos[a+2]=mlight[i].diffuse.z; lightpos[a+3]=mlight[i].diffuse.w;
                       a+=4;
                      }
                      def_shader.uniformmatrixFV("light_dif",4,lightpos);
                      def_shader.off();

                      msaa_shader.reload(0);

                      msaa_shader.on();
                      msaa_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      msaa_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      msaa_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      msaa_shader.uniformI("msaa_smp",msaa_samples);                   // MSAA samples
                      msaa_shader.uniform2F("dxdy",1.0f/(float) (scr_w*msaa_samples), // delta dvoch susediacich texturovacich suradnic
                                                   1.0f/(float) (scr_h*msaa_samples));
                      msaa_shader.off();
                      break;
             case 14: raw_shader.reload(0);

                      raw_shader.on();
                      raw_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      raw_shader.bindattriblocation(2,"in_tan"); // tangenty
                      raw_shader.bindattriblocation(3,"in_bit"); // bitangenty
                      raw_shader.bindattriblocation(4,"in_nor"); // normaly
                      raw_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      raw_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      raw_shader.uniformI("nm_tex_unit",1);      // normalmap texturovacia jednotka
                      raw_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      raw_shader.off();

                      deltas_shader.reload(0);

                      deltas_shader.on();
                      deltas_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      deltas_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      deltas_shader.off();

                      mblur_shader.reload(0);

                      mblur_shader.on();
                      mblur_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      mblur_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      mblur_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      mblur_shader.uniformI("deltas_tex_unit",1);  // texturovacia jednotka s pohybovymi vektormi
                      mblur_shader.uniformI("msaa_smp",msaa_samples);                   // MSAA samples
                      mblur_shader.uniform2F("dxdy",1.0f/(float) (scr_w*msaa_samples), // delta dvoch susediacich texturovacich suradnic
                                                    1.0f/(float) (scr_h*msaa_samples));
                      mblur_shader.off();
                      break;
             case 16: silhouette_shader.reload(1);

                      silhouette_shader.on();
                      silhouette_shader.bindattriblocation(0,"in_pos"); // vrcholy
                      silhouette_shader.bindattriblocation(1,"in_col"); // farby
                      silhouette_shader.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                      silhouette_shader.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                      silhouette_shader.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                      silhouette_shader.off();
                      break;
             default: break;
            }
            break;
   default: break;
  }
  // aktualizacia parametrov osvetlenia
  update_lighting_params();
 }
 else if (but_text_id==10) { // pre HDR shader
  switch (but_id) {
   case 0: hdrparam.refract.x=new_but_data; break;
   case 4: hdrparam.refract.y=new_but_data; break;
   case 8: hdrparam.refract.z=new_but_data; break;
   case 12: hdrparam.fresnel.x=new_but_data; break;
   case 16: hdrparam.fresnel.y=new_but_data; break;
   case 19: hdrparam.fresnel.z=new_but_data; break;
   case 21: hdrparam.bloom.x=new_but_data; break;
   case 22: hdrparam.bloom.y=new_but_data; break;
   case 23: hdrparam.bloom.z=new_but_data; break;
   case 45: // zmena indexu textoveho udaju tlacidla = zmena pouziteho shadera
            but_text_id=(int) new_but_data;
            set_buttons(); // nastavenie parametrov a udajov tlacidiel menu
            // zmena textoveho udaju tlacidla
            but[45].set_data(new_but_data,but_text[but_text_id]);
            if (but_text_id==5) mparam.diffuse.set(0.6f,0.3f,0.1f,1.0f);
            else mparam.diffuse.set(0.8f,0.4f,0.4f,1.0f);
            but[32].set_data(mparam.diffuse.x); but[33].set_data(mparam.diffuse.y);
            but[34].set_data(mparam.diffuse.z); but[35].set_data(mparam.diffuse.w);
            break;
   default: break;
  }
  chromatic_shader.on();
  chromatic_shader.uniformFV("refract",1,hdrparam.refract.get_data()); // refract R, G, B
  chromatic_shader.uniformFV("fresnel",1,hdrparam.fresnel.get_data()); // bias, scale, power
  chromatic_shader.off();

  hdr_shader.on();
  hdr_shader.uniformFV("bloom",1,hdrparam.bloom.get_data()); // factor, bright, exposure
  hdr_shader.off();
 }
 else if (but_text_id==15) { // pre triangle_shrink
  switch (but_id) {
   case 0: shrink_level=new_but_data; break;
   case 4: rotate_speed=new_but_data; break;
   case 45: // zmena indexu textoveho udaju tlacidla = zmena pouziteho shadera
            but_text_id=(int) new_but_data;
            set_buttons(); // nastavenie parametrov a udajov tlacidiel menu
            // zmena textoveho udaju tlacidla
            but[45].set_data(new_but_data,but_text[but_text_id]);
            if (but_text_id==5) mparam.diffuse.set(0.6f,0.3f,0.1f,1.0f);
            else mparam.diffuse.set(0.8f,0.4f,0.4f,1.0f);
            but[32].set_data(mparam.diffuse.x); but[33].set_data(mparam.diffuse.y);
            but[34].set_data(mparam.diffuse.z); but[35].set_data(mparam.diffuse.w);
            break;
   case 46: // reload shadera
            but_id=-1; // reload shadera sa vykona iba 1x
            switch (but_text_id) {
             case 15: triangle_shrink.reload(1);

                       triangle_shrink.on();
                       triangle_shrink.bindattriblocation(0,"in_pos"); // vrcholy
                       triangle_shrink.bindattriblocation(1,"in_col"); // farby
                       triangle_shrink.bindattriblocation(8,"in_uv");  // texturovacie suradnice
                       triangle_shrink.uniformI("base_tex_unit",0);    // zakladna texturovacia jednotka
                       triangle_shrink.uniformmatrixFV("projection_mat",1,avatar.get_pr_mat()); // projekcna matica
                       triangle_shrink.off();
                       break;
             default: break;
            }
            break;
   default: break;
  }
  triangle_shrink.on();
  triangle_shrink.uniform1F("sl",shrink_level);
  triangle_shrink.off();
 }
}

// natocenie svetelneho zdroja *********************************************
void CInterface::light_rotate(GLfloat A,GLfloat B)
{
 lA+=-A; lB+=-B;

 lsrcparam.spotDirection.set(0.0f,0.0f,-1.0f);
 lsrcparam.spotDirection.rotate(lB,rightV);
 lsrcparam.spotDirection.rotate(lA,upV);

 but[16].set_data(lsrcparam.spotDirection.x); but[17].set_data(lsrcparam.spotDirection.y);
 but[18].set_data(lsrcparam.spotDirection.z);
}

// priprava objektu - gula reprezentujuca svetelny zdroj *******************
int CInterface::prepare_light_sphere(float S)
{
 // pomocne premenne
 int e=0;
 int segs=45;
 float alfa[2],beta[2];
 float x[4],y[2],z[4];

 loopj(180) {
  if (j%segs) continue;
  loopi(360) {
   if (i%segs) continue;

   alfa[0]=DEG2RAD((float) i);        beta[0]=DEG2RAD((float) j);
   alfa[1]=DEG2RAD((float) (i+segs)); beta[1]=DEG2RAD((float) (j+segs));

   x[0]=S*cos(alfa[0])*sin(beta[0]); x[1]=S*cos(alfa[1])*sin(beta[0]);
   x[2]=S*cos(alfa[1])*sin(beta[1]); x[3]=S*cos(alfa[0])*sin(beta[1]);
   y[0]=S*cos(beta[0]); y[1]=S*cos(beta[1]);
   z[0]=S*sin(alfa[0])*sin(beta[0]); z[1]=S*sin(alfa[1])*sin(beta[0]);
   z[2]=S*sin(alfa[1])*sin(beta[1]); z[3]=S*sin(alfa[0])*sin(beta[1]);

   sphere_v[e+0]=x[0]; sphere_v[e+1]=y[0]; sphere_v[e+2]=z[0]; // 1. vrchol
   sphere_v[e+3]=x[1]; sphere_v[e+4]=y[0]; sphere_v[e+5]=z[1]; // 2. vrchol
   sphere_v[e+6]=x[2]; sphere_v[e+7]=y[1]; sphere_v[e+8]=z[2]; // 3. vrchol
   e+=9;
   sphere_v[e+0]=x[0]; sphere_v[e+1]=y[0]; sphere_v[e+2]=z[0]; // 1. vrchol
   sphere_v[e+3]=x[2]; sphere_v[e+4]=y[1]; sphere_v[e+5]=z[2]; // 2. vrchol
   sphere_v[e+6]=x[3]; sphere_v[e+7]=y[1]; sphere_v[e+8]=z[3]; // 3. vrchol
   e+=9;
  }
 }

 glGenVertexArrays(1,&sphere_vao);
 glBindVertexArray(sphere_vao);

 sph=new GLbuf;
 if (sph==NULL) {
  printf("Nemozno vytvorit OpenGL zasobnik!\n"); return(1);
 }
 sph->load(sizeof(sphere_v),sphere_v);
 glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,0,0);

 glEnableVertexAttribArray(0);

 glBindVertexArray(0);

 return(0);
}
